from rasa.nlu.components import Component
from rasa.nlu import utils
from rasa.nlu.model import Metadata
import os
import pandas as pd
import typing
from typing import Any, Optional, Text, Dict
import numpy as np
import regex as re
from time import strftime
from datetime import datetime as dt
from typing import Any, Optional, Text, Dict
import numpy as np
import dateutil.parser as dparser
import datefinder

from dateutil.parser import parse
import datetime
import dateutil.parser as dparser
from pyarabic.number import text2number
import datefinder
from dateutil.parser import parse
from datetime import timedelta
from dateutil.relativedelta import relativedelta


from rasa.nlu.training_data import Message, TrainingData


class Extracteur_oncf(Component):
	"""A custom sentiment analysis component"""
	name = "DATA_EXTRACTION"
	provides = ["entities"]
	requires = ["tokens"]
	defaults = {}
	language_list = ["en"]
	print('initialised the class')

	def _init_(self, component_config=None):
		super(Extracteur_oncf, self)._init_(component_config)

	def train(self, training_data, cfg, **kwargs):
		"""Load the sentiment polarity labels from the text
		   file, retrieve training tokens and after formatting
		   data train the classifier."""



	
    
	

	def convert_to_rasa_day(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "date",
				"extractor": "extractor"}

		return entity

	def convert_to_rasa_month(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "mois",
				"extractor": "extractor"}

		return entity
	def convert_to_rasa_horaire(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "tr_horaire",
				"extractor": "extractor"}

		return entity
	
	def convert_to_rasa_numbers(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "number",
				"extractor": "extractor"}

		return entity
	

	
		
	def convert_ville_dep(self, liste):
		"""Convert model output into the Rasa NLU compatible output format."""
		n = len(liste)
		print(n)
		entities= []
		for val in range(n) :
			entities.append({"value" : liste[val],
				  
				"entity": "GARE_DEP",
				"extractor": "extractor"})
		
		return entities
	def convert_gare_dep(self, liste):
		"""Convert model output into the Rasa NLU compatible output format."""
		n = len(liste)
		print(n)
		entities= []
		for val in range(n) :
			entities.append({"value" : liste[val],
				  
				"entity": "GARE_DEP",
				"extractor": "extractor"})

		return entities
	def convert_gare_arr(self, liste):
		"""Convert model output into the Rasa NLU compatible output format."""
		n = len(liste)
		print(n)
		entities= []
		for val in range(n) :
			entities.append({"value" : liste[val],
				  
				"entity": "GARE_ARR",
				"extractor": "extractor"})

		return entities
	def convert_ville_arr(self, liste):
		"""Convert model output into the Rasa NLU compatible output format."""
		n = len(liste)
		print(n)
		entities= []
		for val in range(n) :
			entities.append({"value" : liste[val],
				  
				"entity": "GARE_ARR",
				"extractor": "extractor"})

		return entities

	
	def convert_to_heure(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "HOUR",
				  "extractor": "extractor"}

		return entity


	def convert_to_classe(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "CLASSE",
				  "extractor": "extractor"}

		return entity


	def convert_to_card(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "CARD",
				  "extractor": "extractor"}

		return entity
	
	def convert_to_code_card(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "code_card",
				"extractor": "extractor"}

		return entity


	def process(self, message:Message , **kwargs):
		"""Retrieve the tokens of the new message, pass it to the classifier
			and append prediction results to the message class."""
		print(message.text)	
		if not self :
			entities = []
		else:

			tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)
 
			
			cities_dep = []
			cities_arr = []


#replacing indou arabic numbers
			string = ""
			for t in tokens :
				string = string + t + ' '


			if '۰' in string :
				string = string.replace('۰','0')
			if '١' in string :
				string = string.replace('١','1')
			if '٢' in string :
				string = string.replace('٢','2')
			if '٣' in string :
				string = string.replace('٣','3')
			if '٤' in string :
				string = string.replace('٤','4')
			if '٥' in string :
				string = string.replace('٥','5')
			if '٦' in string :
				string = string.replace('٦','6')
			if '٧' in string :
				string = string.replace('٧','7')
			if '٨' in string :
				string = string.replace('٨','8')
			if '٩' in string :
				string = string.replace('٩','9')
			
			print("string")
			print(string)
			
			tokens = string.split(" ")
			




#detection cities


			#tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)

#delete unecessary de
			for i in range(len(tokens)):
				if tokens[i] == 'de' :
					if tokens[i-1] == 'gare' or tokens[i-1] == 'gares':
						tokens[i] = ""
					
			print('***********new    tokens*****')
			print(tokens)



			datas = pd.read_csv('csv_files/villes_oncf.csv',sep=';',encoding="utf_8")
			Tville= np.array(datas['ville'])
			Oville= np.array(datas['value'])
			ent_val = {}
			city_dep_conv = []
			print("1")
			ville_dep= ""
			ville_arr =""
			entities_city = []
			name_city = []
			print("2")
			for i in range(len(Tville)):
				ent_val[Tville[i]] = Oville[i]
			for word in tokens:
				for key in ent_val:
					if key == word:
						
						name_city.append(word)
						entities_city.append(ent_val[key])
						print(" extracted Token ++++++++++ "+ word)
						break
			print("list city names")
			print(name_city)
			print("list city entities")
			print(entities_city)
			cities_dep = []
			cities_arr = []
			
			if len(entities_city) > 1:
				if tokens[tokens.index(name_city[0])-1] == "من" or tokens[tokens.index(name_city[0])-2] == "من" or tokens[tokens.index(name_city[0])-1] == "de" or tokens[tokens.index(name_city[0])-2] == "de" or tokens[tokens.index(name_city[0])-1] == "depuis" or tokens[tokens.index(name_city[0])-2] == "depuis" or tokens[tokens.index(name_city[0])-2] == "quitter" or tokens[tokens.index(name_city[0])-1] == "quitter":
					ville_dep= entities_city[0]
					ville_arr = entities_city[1]
				elif tokens[tokens.index(name_city[1])-1] == "من" or tokens[tokens.index(name_city[1])-1] == "quitter" or  tokens[tokens.index(name_city[1])-2] == "quitter" or tokens[tokens.index(name_city[1])-2] == "من" or tokens[tokens.index(name_city[1])-1] == "de" or tokens[tokens.index(name_city[1])-2] == "de" or tokens[tokens.index(name_city[1])-1] == "depuis" or tokens[tokens.index(name_city[1])-2] == "depuis"  :
					ville_dep= entities_city[1]
					ville_arr = entities_city[0]
				else :
					ville_dep = entities_city[0]
					ville_arr = entities_city[1]
			if 	len(entities_city) ==1:
				if tokens[tokens.index(name_city[0])-1] == "من" or tokens[tokens.index(name_city[0])-2] == "من" or tokens[tokens.index(name_city[0])-1] == "de" or  tokens[tokens.index(name_city[0])-2] == "de" or tokens[tokens.index(name_city[0])-1] == "quitter" or tokens[tokens.index(name_city[0])-1] == "depuis" or tokens[tokens.index(name_city[0])-2] == "depuis" or tokens[tokens.index(name_city[0])-2] == "quitter" :
					print("yess ville depp")
					ville_dep= entities_city[0]
				else :
					ville_arr = entities_city[0]
			
			
			lt = ville_dep.split(",")
			lt2 = ville_arr.split(",")
			for i in lt :

				cities_dep.append(i)
			for i in lt2 :
				cities_arr.append(i)
			print("villes arrivee")	
			print(cities_arr)
			print("villes de depart")
			print(cities_dep)
			
			


#detection gare	
			
			msg = ""
			for t in tokens :
				if t == "__CLS__" :
					print(" ")
				else :
					msg = msg+" "+t
			
			print(msg+".")
			data_gare = pd.read_csv('csv_files/gare_comp.csv',sep=';',encoding="utf_8")
			exp = np.array(data_gare['gare'])
			gare_nom= np.array(data_gare['value'])
			ent_val = {}
			entity_conv_city = []
			
			entitie_gares = []
			entities_gares = []
			name_gares= []
			entit_gare = ""
			
			for i in range(len(exp)):
				ent_val[exp[i]] = gare_nom[i]
			print(tokens)
			match = ''
			names_word =[]
			for key in ent_val:
				gare = re.search(key,msg)
				
				if gare != None :
					match = gare.group()
					print("MATCH")
					
					entitie_gares.append(ent_val[key])
					name_gares.append(key)
					names_word.append(gare.group())
					break
			match = ''
			print("second gare")
			for key in ent_val:
				
				gare2 = re.search(key,msg)
				
				
				if gare2 != None :
					match = gare2.group()
					print("MATCH2")
					
					
					
				if gare2 != None and ent_val[key] != entitie_gares[0] :
					entitie_gares.append(ent_val[key])
					name_gares.append(key)
					names_word.append(gare2.group())
					break
				else :
					continue
			
			print("gare_names")
			print(name_gares)
			print(names_word)
			
#modiiiiiiiiiiiifiiiiiiiiiiiiicatiooooooooooon
			name_gares = names_word

			
			



			
			
			
			print("entities_gares")
			print(entitie_gares)
			gare1_splitted = []
			gare2_splitted = []
			gare_splitted = []
			gare_dep = []
			gare_arr = []
			
			liste = []
			for i in entitie_gares:
				if i not in liste:
					liste.append(i)
			
			entities_gares = liste
			print("gare_names")
			print(name_gares)
			print("entities_gares")
			print(entities_gares)
			print("entities_city")
			print(entities_city)
			print(cities_arr)
			print(cities_dep)
			print("name_city")
			print(name_city)
			
			print("entitie_gares")
			print(entitie_gares)
			print(entities_city)
			print(cities_dep)
			print(cities_arr)
			
			ind = -1
			ind2 = -1

			for g in entitie_gares :
				if len(entities_city) != 0 :
					if g in entities_city[0]:
							print(entities_city[0])

							ind = 0
					
			print("ok1")
					
			print(entities_city)
			for g in entitie_gares :
				print("ok8")
				print(entities_city)
				if len(entities_city) ==2:
					if g in entities_city[1]:
							print(entities_city[1])
							ind2 = 1
					
				print("ok2")
					
			if ind2 == 1 :
				entities_city.pop(1)
				name_city.pop(1)
			if ind == 0 :
				entities_city.pop(0)
				name_city.pop(0)

			print("entities_city")
			print(entities_city)

			p = 0
			if cities_dep != []:
				for city in cities_dep :
					if city not in entities_city:
						p = 1
						print("1")
						break
				
			z = 0
			if cities_arr != []:
				for city in cities_arr :
					if city not in entities_city:
						z = 1
						print("2")
						break
			
			
			
					

			'''


			index1 = -1
			index2 = -1
			cities_dep_new = []
			cities_dep_text = ''

			
					if city in entities_city[0]:
						
						index1 = 0
						break
				print("indew1")
				print(index1)
				for city in cities_dep :
					if city in entities_city[1]:
						index1 = 1
						break
				print("indew1")
				print(index1)

			print(entities_city)
			if cities_arr != []:
				for city in cities_arr :
					if city in entities_city[0]:
						
						index2 = 0
						break
				print("indeww1")
				print(index2)
						
				for city in cities_arr :
					if city in entities_city[1]:
						index2 = 1
						break
				print("indeww2")
				print(index2)

			print(index1)
			print(index2)
			print(entities_city)

			if index1 == 0 or index2 == 0:
				entities_city.pop(0)
			if index1 == 1 or index2 == 1:
				entities_city.pop(1)
			print("new entities_city")
			print(entities_city)
			
			if cities_dep == [] :
				entities_city.pop(0)
			if cities_arr == [] :
			entities_city.pop(1)

			entities_city=[]
			if cities_dep != []:

				entities_city.append(cities_dep)
			if cities_arr != []:

				entities_city.append(cities_arr)
			

			


			cities_det1 = []
			var = 0

			if len(entities_city) != 0 and entities_city[0] != [] :
				cities_det1 = entities_city[0].split(",")
				print(cities_det1)
			if len(entities_city) != 0 and entities_city[0] != [] :
				cities_det1 = entities_city[0].split(",")
				print(cities_det1)
			
			
			
			for gare in entitie_gares :
				if gare in cities_det1:
					var = 1
			
			print("var")
			print(var)
			
			if len(entities_gares) == 1 and len(entities_city) == 1 and var == 1:
						print("yes entitycity == entity gare")
						entities_city = []
						cities_arr = []
						cities_dep = []
						print(entities_city)
						print(entities_gares)

						print()
		'''
			if len(entities_gares) == 2 :
				if entities_gares[1] == entities_gares[0]:
					entities_gares[1] == ''
			print("ok")
			if len(entities_gares) == 1 : 
				if len(entities_city) == 0 :
					print("ok1")
					if " من" in msg or  "de " in msg or  "depuis " in msg or "quitter" in msg:
							print("depuiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiis")
							gare_dep.append(str(entities_gares[0]))
					if gare_dep == []:
						gare_arr.append(str(entities_gares[0]))
				elif len(entities_city) == 1 :
					print("ok 1 ville 1 gare")
					
					print(entities_gares)
					print(entities_city)
					gare_splitted = name_gares[0].split(" ")
					print(gare_splitted)
					print("ok2")
#CHECKER SI LA GARE DETECTEE EST UNE GARE DE DEP OU PAS
					print(tokens[tokens.index(gare_splitted[0])])
					print("ok3")
					print(name_city[0])
					if tokens[tokens.index(gare_splitted[0])-1] == 'de' or tokens[tokens.index(gare_splitted[0])-1] == 'من' or tokens[tokens.index(gare_splitted[0])-1] == 'depuis' or tokens[tokens.index(gare_splitted[0])-1] == 'quitter':
						print("depuiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiis")
						gare_dep.append(str(entities_gares[0]))
					elif tokens[tokens.index(name_city[0])-1] == 'de' or tokens[tokens.index(name_city[0])-1] == 'من' or tokens[tokens.index(name_city[0])-1] == 'depuis' or tokens[tokens.index(name_city[0])-1] == 'quitter':
						gare_arr.append(str(entities_gares[0]))
						cities_dep= entities_city[0].split(",")
					else :
						print("non depuiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiis")
						k = tokens.index(name_city[0])
						j = tokens.index(gare_splitted[0])
##MODIIIIFIICAAATIOOON
						cities_dep = []
						cities_arr = []
						print(k)
						print(j)
#cas (casa port)
						if j == k :
									cities_arr = []
									print("alike")
									gare_arr.append(str(entities_gares[0]))
									print(gare_arr)
#cas (fes casa port)
						elif j > k :
									gare_arr.append(str(entities_gares[0]))
									cities_dep= entities_city[0].split(",")
									print('j > k')
									print(gare_arr)
									print(cities_dep)
# cas (casa port fes)
						else : 
									gare_dep.append(str(entities_gares[0]))
									cities_arr= entities_city[0].split(",")
									print('k > j')
									print(gare_dep)
									print(cities_arr)
#2 VILLES ET UNE GARE , l'une des villes doit etre enlevee
				elif len(entities_city) == 2 :
					gare_splitted = name_gares[0].split(" ")
					print("ok 2 ville 1 gare")
					print(entities_gares)
					print(entities_city)
					k = tokens.index(name_city[0])
					print(k)
					l = tokens.index(name_city[1])
					print(l)
					j = tokens.index(gare_splitted[0])
					print(j)
					if j == k :
									phr = entities_city[1].split(",")
									print(phr)

									print(entities_gares)
									print(entities_city)
									print('k===j')
									name_city[0] = ''
									print(name_city)
									if j > l :
										print("j suppp")
										print(entities_city[1])
										for i in phr :
											cities_dep.append(i)
										gare_arr.append(entities_gares[0])
									
										print(cities_dep)
										print(gare_arr)
									if l> j : 
										print("l suupp")
										print(entities_city[1])
										
										if tokens[l-1] == 'de' or tokens[l-1] == 'من' or tokens[l-1] == 'depuis' or tokens[l-1] == 'quitter':

											gare_arr.append(entities_gares[0])
											for p in phr:
												cities_dep.append(p)
										else :
											for i in phr :
												cities_arr.append(i)
											gare_dep.append(entities_gares[0])
										print("cities_arr")
										print(cities_arr)
										print("gare_dep")
										print(gare_dep)
					if j == l :
									phr = entities_city[0].split(",")
									print(phr)
									print("j==l")
									name_city[1] = ''
									if j > k :
										if tokens[j-1] == 'de' or tokens[j-1] == 'من' or tokens[j-1] == 'depuis' or tokens[j-1] == 'quitter' :

											gare_dep.append(entities_gares[0])
											for p in phr:
												cities_arr.append(p)


										else :
											for i in phr :

												cities_dep.append(i)
										
											gare_arr.append(entities_gares[0])
									if k > j : 
										for i in phr :
											cities_arr.append(i)
										gare_dep.append(entities_gares[0])
					

					if len(cities_dep) == 2 :
								if cities_dep[0] == cities_dep[1]:
									cities_dep.remove(cities_dep[1])
					if len(cities_arr) == 2 :
								
								if cities_arr[0] == cities_arr[1]:
									cities_arr.remove(cities_arr[1])
					cities_arr = list(dict.fromkeys(cities_arr))
					cities_dep = list(dict.fromkeys(cities_dep))
					print("okkkkkkkkkkkkkkkkkkkkkkk")
					print(cities_dep)
					print(cities_arr)
					print(gare_dep)
					print(gare_arr)
			

#le cas de 2 gares
			print("entities_city")
			print(entities_city)
			print("cities_arr")
			print(cities_arr)
			print("cities_dep")
			print(cities_dep)
			
			if len(entities_gares) > 1 and entities_gares[0] != entities_gares[1] :
				print("2 gares")
				print(name_gares)
				gare1_splitted = name_gares[0].split(" ")
				gare2_splitted = name_gares[1].split(" ")
				print("gare1_splitted")
				print(gare1_splitted)
				print("gare2_splitted")
				print(gare2_splitted)
				i1 = msg.find(gare1_splitted[0])
				i2 = msg.find(gare2_splitted[0])
				if i1 > i2 :
					m = gare1_splitted
					gare1_splitted = gare2_splitted
					gare2_splitted = m
					p = entities_gares[0]
					entities_gares[0] =entities_gares[1]
					entities_gares[1] = p
				
				if i2 > i1 :
					print("no")
				if i2 == i1 :
					if gare1_splitted[1] != "" and gare2_splitted[1] != "" :
						i3 = msg.find(gare1_splitted[1])
						i4 = msg.find(gare2_splitted[1])
						if i3 > i4 :
							m = gare1_splitted
							gare1_splitted = gare2_splitted
							gare2_splitted = m
							p = entities_gares[0]
							entities_gares[0] =entities_gares[1]
							entities_gares[1] = p
				print("second")
				print("gare1_splitted")
				print(gare1_splitted)
				print("gare2_splitted")
				print(gare2_splitted)
#take the phrase before both station names
				pre_gare1 = ""
				for i in range(tokens.index(gare1_splitted[0])):
					pre_gare1 = pre_gare1 + " " + tokens[i]
				print("pre_gare1")
				print(pre_gare1)
				pre_gare2 = ""
				gare1splitted_fin = ''
				if len(gare1_splitted) == 1 :
					print("ok1")
					gare1splitted_fin = gare1_splitted[0]
					print(gare1splitted_fin)
					print(gare1_splitted)
					print(gare1_splitted[0])
				if len(gare1_splitted) == 2 :
					print("ok2")
					
					gare1splitted_fin = gare1_splitted[1]
					print(gare1splitted_fin)
				print("okk")
				for i in range(tokens.index(gare1splitted_fin),tokens.index(gare2_splitted[0])):
					print(i)
					pre_gare2 = pre_gare2 + " " + tokens[i]
					
				print("pre_gare2")
				print(pre_gare2)
				

			

				if  "من" in pre_gare1 or  "de" in pre_gare1 or "depuis" in pre_gare1 or  "quitter" in pre_gare1:
					print("de in pregare1")
					gare_dep.append(str(entities_gares[0]))
					print(str(entities_gares[1]))
					gare_arr.append(str(entities_gares[1]))
					print(str(entities_gares[0]))
				elif  "من" in pre_gare2 or  "de" in pre_gare2 or  "depuis" in pre_gare2 or  "quitter" in pre_gare2:
					print("de in pregare2")
					gare_dep.append(str(entities_gares[1]))
					print(str(entities_gares[1]))
					gare_arr.append(str(entities_gares[0]))
					print(str(entities_gares[0]))
				else :
					gare_dep.append(str(entities_gares[0]))
					gare_arr.append(str(entities_gares[1]))
			print("gare_dep")
			print(gare_dep)
			print("gare_arrr")
			print(gare_arr)
			
			cities_arr = list(dict.fromkeys(cities_arr))
			cities_dep = list(dict.fromkeys(cities_dep))

			print('cities_arr')
			print(cities_arr)
			print("cities_dep")
			print(cities_dep)
			
			
			
			
			
			
			
			
			
									
									



			



#Detection Date
			stri = ""
			now = dt.now() 
			date = str(now).split(" ")
				
			now_day_1 = now - timedelta(days=now.weekday())
			current_year = dt.now().year

			dates = {}
			entit_jour = ""
			entit_mois = ""
#detection du jour de la semaine
			for n_week in range(1):
				dates[n_week] = [(now_day_1 + timedelta(days=d+n_week*7)).strftime("%m/%d/%Y") for d in range(7)]
			print(dates)
				
			datas_jour = pd.read_csv('csv_files/jour_dict.csv',sep=';',encoding="utf_8") 
			jour= np.array(datas_jour['jour'])
			print("ok")
			value_jour= np.array(datas_jour['value'])
			print("ok")
			ent_jour = {}
			for i in range(len(jour)):
				ent_jour[jour[i]] = value_jour[i]
			
			stri = ''
			for t in tokens :
				stri = stri + ' ' + t
			print(stri)
			print(match)
			if match != '' :
				toks = stri.replace(match,' ')
			else :
				toks = tokens
			print(toks)
			if isinstance(toks, list):
				print("array")
				tok = toks
			else :
				tok = toks.split(' ')
			print(tok)
			
				
			for word in tok:
				print(word)
				for key in ent_jour:
					if key == word:
						
						print("ok3")
						today_n = datetime.datetime.today().weekday()
						print(today_n)
						print(ent_jour[key])
						if int(ent_jour[key]) < today_n or int(ent_jour[key]) == today_n:
							print("prochain jouuuuuur")
							print(today_n)
							print(ent_jour[key])
							entite_jour = dates[0][int(ent_jour[key])]
							datetimeobject = dt.strptime(entite_jour ,'%m/%d/%Y')
							entit_jour  = datetimeobject.strftime('%d/%m/%Y')
							date_1 = dt.strptime(entit_jour, "%d/%m/%Y")
							date_next_week = date_1 + relativedelta(days=7)
							print(date_next_week)
							date_next = str(date_next_week).split(' ')[0]

							datetimeobject = dt.strptime(date_next ,'%Y-%m-%d')
							entit_jour  = datetimeobject.strftime('%d/%m/%Y')

						else :
							print('jouur de la semaiiine')
							entite_jour = dates[0][int(ent_jour[key])]
							datetimeobject = dt.strptime(entite_jour ,'%m/%d/%Y')
							entit_jour  = datetimeobject.strftime('%d/%m/%Y')
							print(entit_jour)



			
			
					
				
			
						
# "prochain" 
			print("apres midi")
			print(tokens)
			indexs = []
			for j in range(0,len(tokens)):
				if tokens[j] == "après" or tokens[j] == "apres" or tokens[j] == "بعد" or tokens[j] == "باعد":
					indexs.append(j)
					print(j)
					print("apres")
			
			for ind in indexs :
				if  tokens[ind+1] == "midi" or tokens[ind+1] == "الزوال" or tokens[ind+1] == "الظهيرة" or tokens[ind+1] == "زوالا" :
						print("midi")
						print(tokens[j])
						del tokens[ind]


			print(tokens)
			proch = ['prochain','prochaine','suivant','suivante','après','apres','جاي','الجاي','الماجي','الجايا','جايا','الماجيا',"بعد","باعد","جيي"]
			prochain = 0
			for t in tokens :
					if t in proch :
						prochain = 1


			if entit_jour != "" :
				det_date = entit_jour.split("/")
				j = det_date[0]
				m = det_date[1]
				print(j)
				print(m)
				date_jour = str(now).split(" ")
				
				datetimeobjecte = dt.strptime(date_jour[0] ,'%Y-%m-%d')
				jour_t  = datetimeobjecte.strftime('%d/%m/%Y')
				detect_date = jour_t.split("/")
				j_auj = detect_date[0]
				m_auj = detect_date[1]
				print(m_auj)
				print(j_auj)
				
				if int(m_auj) == int(m):
					if int(j_auj) > int(j) :
						print("lundi next week")
						date_2 = dt.strptime(entit_jour, "%d/%m/%Y")
						date_next_w = date_2 + relativedelta(days=7)
						print(date_next_w)
						date_nxt = str(date_next_w).split(' ')[0]
						datetimeobjectee = dt.strptime(date_nxt ,'%Y-%m-%d')
						entit_jour = datetimeobjectee.strftime('%d/%m/%Y')

			

					
						
			print('entitjouur'+entit_jour) 

#detection d'aujourdhui et lyoum		
						
			
				
			if entit_jour == "":
				date = str(now).split(" ")
				print("today")
				print(date)
				for word in tokens: 
							if  word == "ليوم" or word == "aujourd" or word == "aujourdhui" or word == "maintenant" or word == "دابا" or word == "اليوم" or word == "الان" or word == "aujourd'hui":
								entit_jours = date[0]
								datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
							if word == "غدا" or word== "demain" or word== 'غادا' or word =="غادان" or word== "غد" or word == "دماين" or word == "دومان" or word == "غحدا" or word == "dem1" or word== "دوماين" or word =="دمان":
								if prochain == 0 :
									entit_jours = str(datetime.date.today() + datetime.timedelta(days=1))
									datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
									entit_jour  = datetimeobject.strftime('%d/%m/%Y')
									print("*********************1")
									print(str(datetime.date.today() + datetime.timedelta(days=1)))
								if prochain == 1 :
									entit_jours = str(datetime.date.today() + datetime.timedelta(days=2))
									datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
									entit_jour  = datetimeobject.strftime('%d/%m/%Y')
									print("*********************2")
									print(str(datetime.date.today() + datetime.timedelta(days=2)))
							if word== "lendemain":
								entit_jours = str(datetime.date.today() + datetime.timedelta(days=2))
								datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
								print("*********************")
								print(str(datetime.date.today() + datetime.timedelta(days=2)))
							if word== "surlendemain":
								entit_jours = str(datetime.date.today() + datetime.timedelta(days=3))
								datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
								print("*********************")
								print(str(datetime.date.today() + datetime.timedelta(days=3)))
							if word== "hier" or word == "لبارح" or word == "الامس":
								entit_jours = str(datetime.date.today() - datetime.timedelta(days=1))
								datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
								print("*********************")
								print(str(datetime.date.today() - datetime.timedelta(days=1)))
#detection des formats


#format jj/mm/yyyy jj.MM
			list_words = tokens
			
			if entit_jour == "" :
				for i in range(len(list_words)):
					if list_words[i] == "هاد" or list_words[i] == "هذا" :
						if list_words[i+1] == "نهار" or list_words[i+1] == "النهار" :
							entit_jours = date[0]
							datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
							entit_jour  = datetimeobject.strftime('%d/%m/%Y')

			if entit_jour == "": 
					for word in tokens:
						dates = re.search(r'(\d{1,2}([/])\d{1,2}([/])\d{1,4})',word)
						date7 = re.search(r'(\d{1,2}([.])\d{1,2})',word)
						if dates == None :
							print("----------------------------------")
							if date7 != None :
								print("date......... found")
								datetimeobject = dt.strptime(date7.string ,'%d.%m.%Y') 
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
						else :	
							entit_jour = dates.string 
			print("entit_jour***************************************************")
			print(entit_jour)
			if entit_jour.endswith('/21'):
				print('/21')
				entit_jour =entit_jour.replace("/21","/2021")
			elif entit_jour.endswith('/22'):
				entit_jour =entit_jour.replace("/22","/2022")
			print(entit_jour)
			tok = ""
			for t in tokens :
				tok = tok + " " + t
#DETECTION DE JJ/MM
			dates1 = re.search(r'(\d{1,2}([/])\d{1,2})',tok)
			print('okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk')
			if entit_jour == "":
							print("nooooooooooooooooooooooooooooo")
							if dates1 == None :
								print('nooooooooooooooooooooooooooooooooooooooooooooooooooooooooo')
							else :
								print("ooooooooooooooooooooooooooooooooooooooooooooooooooooooook")
								entity_jour  = dates1.group()
								print(dates1.group())
								print("entity_jour")
								print(entity_jour)
								listt = entity_jour.split("/")
								j=listt[0]
								m=listt[1]
								print("JJ/MM")
								print(entity_jour)
								
								entit_jour = entity_jour +"/" +str(current_year)
								
#DETECTION DE JJ-MM-YYYY

			datess = re.search(r'(\d{1,2}([-])\d{1,2}([-])\d{1,4})',tok)
			print('okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk')
			if entit_jour == "":
							print("nooooooooooooooooooooooooooooo")
							if datess == None :
								print('nooooooooooooooooooooooooooooooooooooooooooooooooooooooooo')
							else :
								print("ooooooooooooooooooooooooooooooooooooooooooooooooooooooook")
								entity_jour  = datess.group()
								tt = datess.group().split()
								print(tt[0])
								datetimeobject = dt.strptime(tt[0] ,'%d-%m-%Y')
								print(datetimeobject)
								entit_jour  = datetimeobject.strftime('%d/%m/%Y')
#DETECTION DE JJ-MM		
							dates2 = re.search(r'(\d{1,2}([-])\d{1,2})',tok)
							if dates2 != None :
								print("----------------------------------")
								tt = dates2.group().split()
								print(tt[0])
								datetimeobject = dt.strptime(tt[0],'%d-%m')
								print(datetimeobject)
								entit_jour_sans_annee  = datetimeobject.strftime('%d/%m')
								listt = entit_jour_sans_annee.split("/")
								j=listt[0]
								m=listt[1]
								entit_jour = entit_jour_sans_annee  + "/" +str(current_year)
								
								print(entit_jour)
							else : 
								print("no")
			#DETECTETION DE 06 12 COMME DATE
							'''dates3 = re.search(r'(\d{1,2}.\d{1,2})',tok)
							print(tok)
							if dates3 != None :
								print('date3')
								tt = dates3.group().split()
								text = tt[0] +' ' +tt[1]
								datetimeobject = dt.strptime(text,'%d %m')
								print(datetimeobject)
								entit_jour_sans_anne  = datetimeobject.strftime('%d/%m')
								entit_jour = entit_jour_sans_anne + "/2020"
								print(entit_jour)

							'''	

			tok = ""
			for t in tokens :
				tok = tok + " " + t
#DETECTION DE JJ MM ET JJ MM 2020

			date3 = re.search(r'(\d{1,2}([\s])\d{1,2})',tok)
				
			l = []
			bb = ""
			if entit_jour == "":
							print('date full-------------------------------------------------------------------------')
							if date3 != None :
								print('date full-------------------------------------------------------------------------')
								entitee_jour = date3.string
								print(entitee_jour)
								entitees_jour= entitee_jour.split(" ")
								print(entitees_jour)
								for i in entitees_jour :
									if i.isdigit() :
										if  i!= "2021":
											print('i is digit')
											l.append(i)
											l.append("/")
										else :
											l.append(i)
									else :
										print("CLS")
								print(l)
								for t in l :
									if t != "2021" :
										bb = bb + t
								entiti_jour = bb 
								print(entiti_jour)
								listt = entiti_jour.split("/")
								j=listt[0]
								m=listt[1]
								
								entit_jour = entiti_jour   +str(current_year)

			print("entite_jour"+entit_jour)

		
## Extractor mois :
			print("month detection :::::::")
				#detect month written in arabic words 	
			datas = pd.read_csv('csv_files/mois_dict.csv',sep=';',encoding="utf_8") 
			mois= np.array(datas['mois'])
				
			value= np.array(datas['value'])
			ent_mois = {}

			for i in range(len(mois)):
					ent_mois[mois[i]] = value[i]

			if entit_mois == "" :
				
					for word in tokens:
						for key in ent_mois:
							if key == word:
							
								entit_mois = str(ent_mois[key])
				
			if entit_mois == "" :
					print("there is no month !!!!!!")
			else : 
					print("***********************month in arabic words"+entit_mois)
				# detect month
				
			n = 0
			nb = ""
	#detect the word chhr in the sentence and take the following word if entit_mois is not already full
			
			nbre_ar = pd.read_csv('csv_files/nbre_ar.csv',sep=';',encoding="utf_8") 
			nbre= np.array(nbre_ar['nbre'])
				
			value= np.array(nbre_ar['value'])
			ent_nbre = {}
			month_arab = 0
			for i in range(len(nbre)):
					ent_nbre[nbre[i]] = value[i]
			if entit_mois == "":
					month_arab=0
					for j in range(len(tokens)):
						if tokens[j] == "شهر" or tokens[j] == "الشهر" or tokens[j] =="شحر" :
								n = j
								month_arab=1
					nb = tokens[n+1]
		#detect month after the word 'month' written in digital numbers 
					#for u in range(n,len(tokens)):
					#	reste= reste +" " + tokens[u]
					print("nb"+ nb)
					if nb.isdigit() and int(nb) < 13:
						entit_mois = nb
					else: 
						print("no digital ")
						for key in ent_nbre :
							if key == nb :
								entit_mois = str(ent_nbre[key])
								print("october")
			mois_fr = -2
			if entit_mois == "" :
				for j in range(len(tokens)):
						if tokens[j] == "mois" :
								mois_fr= j
								print(mois_fr)
								print("----------------------------------------")
								print(tokens[mois_fr+1])
				if tokens[mois_fr+1].isdigit() and int(tokens[mois_fr+1]) < 13:
						entit_mois = tokens[mois_fr+1]
		#detect month after 'month' written in arabic numbers
			if entit_mois == '':
						print("*******************************************************still no month")
						conv_mon = text2number(nb)
						if conv_mon != 0:
							print("converted month number")
							print(conv_mon)
							entit_mois = conv_mon
						else :
							print('no month in arabic numbers')
			else : 
						print("********************************************************month"+entit_mois)


			current_month = dt.now().month
			current_year = dt.now().year
			print(entit_jour)
			print("okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk")
			
				
		
	## Extract number
			entity_numb = ""
			for t in tokens :
				if t.isdigit() and t!= "2020" and t != '2021':
					entity_numb = t
			

			text1= ''
			texte= ''
			if n == 0 :
				text1 = text1 + " " + tokens[0]
			else : 
				for i in range(0,n):
					text1 = text1 + " " + tokens[i]
			for i in range(0,len(tokens)):
					texte = texte + " " + tokens[i]
			print("text1"+str(n))
			print(tokens[0])
			print(text1) 
			
			list_numb = []
			ph = ''
			texte11 = text1.split(" ")
			print(texte11)
	#detect day if before word chhr and digit
			for word in texte11:
					ph = ph + " " + word
					if word.isdigit() :
						print("++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"+word)
						if word != '2020' and word != "2021":
							entity_numb = word
			if len(entity_numb) == 1 :
					entity_numb = "0"+entity_numb
				
	#detect day if before word chhr and written in arabic words
			print(entit_jour)
			print("okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk")
			conv_num = text2number(text1)
			print(conv_num)
			if conv_num == 0 :
					print("no")
			else : 
					entity_numb = conv_num
			print(entity_numb)
			print("entitynumb")
			print("------------------------------list nombres en arabe")
			print(entity_numb)
			#if month_arab == 0 :
					#conv_num = text2number(texte)
					#if conv_num == 0 :
						#print("no arabic number")
					#else :
						#if conv_num < 32 :
							#entity_numb = conv_num
			print("entitynumbbb")
			print(entity_numb)
#dans 2 jours / dans deux jours
			nbre_ar = pd.read_csv('csv_files/nbre_ar.csv',sep=';',encoding="utf_8") 
			nbre= np.array(nbre_ar['nbre'])
				 
			value= np.array(nbre_ar['value'])
			ent_nbre = {}
			for i in range(len(nbre)):
						ent_nbre[nbre[i]] = value[i]
			key_numb= ''
			key_val = ''
			for w in tokens:
					for key in ent_nbre:
							if key == w :
								print('keeeeeeeeeey')
								print(w)
								print(key)
								key_numb = key
								key_val = ent_nbre[key]
			print(key_numb)
			print(key_val)
			if entit_jour == "":
				d = -1
				for j in range(len(tokens)):
					if tokens[j] =="jour" or tokens[j] =='jours' or tokens[j] == 'أيام' or tokens[j] == 'يوم':
						d = j

				if tokens[d-1].isdigit():
					dig = tokens[d-1]
				
					entit_jours = str(datetime.date.today() + datetime.timedelta(days=int(dig)))
					datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
					entit_jour  = datetimeobject.strftime('%d/%m/%Y')


				print(tokens[d-1])

				if tokens[d-1] == key_numb :
					print('d-1')
					dig = key_val
					entit_jours = str(datetime.date.today() + datetime.timedelta(days=int(dig)))
					datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
					entit_jour  = datetimeobject.strftime('%d/%m/%Y')
			print("dans queeelqueeees jouuurs")
			print(entit_jour)
			print(key_numb)
			

				
			nbre_ar = pd.read_csv('csv_files/nbre_ar.csv',sep=';',encoding="utf_8") 
			nbre= np.array(nbre_ar['nbre'])
				 
			value= np.array(nbre_ar['value'])
			ent_nbre = {}
			for i in range(len(nbre)):
						ent_nbre[nbre[i]] = value[i]
			if entity_numb == 0:
					print("entitynumber 0")
					nbr = ""
					for w in tokens:
						for key in ent_nbre:
							if key == w :
								print("w")
								print(w)
								entity_numb = ent_nbre[key]
#if entit_mois is more than 12 
			print("yeeeeeeeeeeeeeeeees heeeeeeeeeeere")
			print(entity_numb)
			#if entit_mois != "" :
				#if int(entit_mois) > 12 :
					#entit_mois = ""
			if len(str(entity_numb)) == 1 :
				entity_numb = "0" + str(entity_numb)

#entit_mois or entity_numb 
			if entit_jour == "":
					if entity_numb != "":
						print(entity_numb)
						if entit_mois != "" :
							print("yes month")
							print(current_month)
							print(entit_mois)
							
							entit_jour = str(entity_numb) + "/" + str(entit_mois) +"/" +str(current_year)
						else :
							entit_jour = str(entity_numb) + "/" + str(current_month) +"/" +str(current_year)
							
						
						
			if len(entit_jour) > 11 :
				entit_jour = ""	
			if entity_numb == "" :
				for word in tokens:
					if word.isdigit() :
						print("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++entit_number"+word)
						if word != '2020' and word != "2021":
							entity_numb = word
			if entity_numb == "" :
				for word in tokens:
					for key in ent_nbre:
							if key == word :
								entity_numb = ent_nbre[key]

			if entit_jour =="" :
				nhar = ''
				ch = ''
				for i in range(len(tokens)):
						if tokens[i] == "نهار" or tokens[i] == "يوم" :
									
							nhar = tokens[i+1]
							print("nhaaaar"+ nhar)
							ch = tokens[i+2]
				if nhar.isdigit() and entit_mois != "":
					entit_jour = nhar +"/" + entit_mois +"/" +str(current_year)
				if nhar.isdigit() and entit_mois == "":
					entit_jour = nhar +"/" + str(current_month) +"/" +str(current_year)
				if nhar.isdigit() and ch.isdigit():
					entit_jour = nhar +"/" + ch +"/" +str(current_year)

			entit_jours = date[0]
			datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
			auj  = datetimeobject.strftime('%d/%m/%Y')
			print(entit_jour+"entitjouuuuur")
			print(auj)


#detect moments of the day


			terms = ['عشيه','لعشيا','ليل','soir','midi','matin','العشيه']
			if entit_jour == "" :
				for t in tokens :
					if t in terms :
						entit_jours = date[0]
						datetimeobject = dt.strptime(entit_jours ,'%Y-%m-%d')
						entit_jour  = datetimeobject.strftime('%d/%m/%Y')
			
			

#delete wrong dates with wrong months

			'''if entit_jour != "":
				det_date = entit_jour.split("/")
				mon = det_date[1]
				jouur = det_date[0]
				print(mon)
				if int(mon) > 12 :
					entit_jour = ""
				if int(jouur) > 31 :
					entit_jour = ""
			'''

#delete date anterieurs :
			

			numbers = []
			aujourdhui = str(now).split(" ")
#جوج شهر خمسة
			numbers_ar = []
			if entit_jour == '':
				for t in tokens :
					for key in ent_nbre:
							if t == key :
								numbers_ar.append(ent_nbre[key])
				print("numbers_ar")
				print(numbers_ar)
			
				for t in toks :
					if t.isdigit():
						numbers.append(t)
				print("numbeeers")
				print(numbers)

			




			print("Hour detection :::::::")
			toks = [t.text for t in message.get("tokens")]
			heure = ""
	#Extract regular formats 
			for t in toks:
					print(t)
					hour1=re.search(r'(([0-9]+)+(h|heure|heures).([0-9]+))', t)
					if hour1 == None :
						print("no hour(h)")
						hour1 = re.search(r'(([0-9]+)+(h|heure|heures))',t)
						if hour1 == None :
							
							hour1 = re.search(r'(([0-9]+):([0-9]+))', t)
							if hour1 == None :
								print("no hour(:) detected in extractor")
							else :
								heure = hour1.string
								if len(heure) < 5:
									heure = "0"+heure
								
						else :
							heure1 = hour1.string
							for i in heure1 :
								print(i)
								if i.isdigit():
									heure += i
							heure = heure +":00"
							if len(heure) < 5:
									heure = "0"+heure

					else :
						heure1 = hour1.string
						for i in heure1 :
								print(i)
								if i.isdigit():
									heure += i
								if i == "h" :
									heure = heure + ":"
									continue
						if len(heure) < 5:
									heure = "0"+heure

						print("heur(h) found")
			h=-1






	#Extract heure after sa3a and ma3a 


			if heure == "" :
				for i in range(len(toks)) : 
					if toks[i] == "الساعة" or toks[i] == "ساعة" or toks[i] == "مع" or toks[i]== "معا" or toks[i] == "الساعه" or toks[i] == "ساعه" or toks[i] == "vers":
						h= i
			print(toks[h+1])
			if h == -1 :
				print("no heure")
			else :
				print("yeeeeeeeeeeeees heure !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
				if toks[h+1].isdigit():
					heure = toks[h+1]
					print(heure)
					print('yeeeeeeeeeeeees heure !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
				else :
					for key in ent_nbre:
							if key == toks[h+1] :
								print("//////////////////////////////////////////////////////////////////////////////////"+heure)
								heure = ent_nbre[key]



	#Extract heure with minutes written in darija 

			dat = pd.read_csv('csv_files/minutes.csv',sep=';',encoding="utf_8") 
			mins= np.array(dat['mins'])
				
			vale= np.array(dat['value'])
			ent_min = {}
			for o in range(len(dat)):
				ent_min[mins[o]] = vale[o]
			dats = pd.read_csv('csv_files/moins_minutes.csv',sep=';',encoding="utf_8") 
			mi= np.array(dats['min'])
			vals= np.array(dats['value'])
			ent_moins_mins = {}
			for o in range(len(dat)):
						ent_moins_mins[mi[o]] = vals[o]
			moins = 0 
			for t in toks : 
				if t == 'قل':
					moins = 1
			hh = [0]
			if len(heure)> 2 :
				hh = heure.split(":")
				h = hh[0]
			else :
				h = 0
			print(h)
			print(hh)
			if heure != "" and int(h) < 19:
				print("111111111111111111111111111111111111111111")
				for t in toks :
					for key in ent_min:
						if key == t and moins == 0 :
								if len(str(ent_min[key]))!= 1 :
									heure = heure + ':' + str(ent_min[key])
									#print("ok2")

								else :
									minute = str(ent_min[key])

									heure = heure + ':0' + str(ent_min[key])
			print(heure)	
			if heure != "" and int(h) < 19:
				for t in toks :
					for key in ent_moins_mins:
							if key == t and moins == 1 :
								if len(str(ent_moins_mins[key]))!= 1 :
									minute = str(ent_moins_mins[key])
									heure = heure + ':' + str(ent_moins_mins[key])

								else :
									minute = str(ent_moins_mins[key])

									heure = heure + ':0' + str(ent_moins_mins[key])
			print("heur"+heure)


#detect hour from entity_number

			if entity_numb != "" and heure == "" and int(entity_numb) < 24  :
				#print(moins)
				for t in toks :
					#print(t)
					for key in ent_min:
						if moins == 0 :
							if key == t  :
								print("ok4")
								if len(str(ent_min[key]))!= 1 :
									minute = str(ent_min[key])
									heure = entity_numb+ ':' + str(ent_min[key])

								else :
									minute = str(ent_min[key])

									heure = entity_numb + ':0' + str(ent_min[key])
			if entity_numb != "" and heure == "" :
				for t in toks :
					for key in ent_moins_mins:
							if key == t and moins == 1 :
								print("ok5")
								if len(str(ent_moins_mins[key]))!= 1 :
									minute = str(ent_moins_mins[key])
									heure = entity_numb + ':' + str(ent_moins_mins[key])

								else :
									minute = str(ent_moins_mins[key])

									heure = entity_numb + ':0' + str(ent_moins_mins[key])
			d = 0
			k = -1
			if heure.isdigit():
				if int(h) < 19 :
					if len(heure) == 1 or len(heure) ==2 :
						for tt in toks :
							if tt.isdigit():
								print("lol")
								k = tt
								print(k)
								d = 1
			print('last ok')
			print(heure)
			if k != -1 :
				if k != heure :
					heure = heure + ':' + k 
				else :
					heure = heure + ':00'
			if len(heure)== 3 :
				if entity_numb!= "" and int(entity_numb) < 19 and entity_numb != heure:
					heure =  heure + entity_numb  
			print(entity_numb)
			print(heure)
			if d == 0 and len(heure) < 3 and heure != "":
				heure = heure + ":00"
			
	#Extract heure with minutes sous forme daqiqa 
			numbers = []
			for i in toks :
				if i.isdigit():
					numbers.append(i)
			phrase = ""
			for t in tokens :
				phrase = phrase + " " + t
			numb = ""
			for word in tokens:
					for key in ent_nbre:
							if key == word :
								numb = ent_nbre[key]
			print(numbers)
			print(len(numbers))
			minu = 0
			print('[((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((')
			if len(numbers) == 1 or len(numbers) == 2   :
				print("yes")
				if  heure == "":
					print("55555")
					minut = "00"
					for j in range (len(toks)) :
						if toks[j] == 'دقيقة' or toks[j] == "دقيقه" :
							minut = toks[j-1]
							minu = 1
					if minut.isdigit() :
						if numbers[0] != minut :
							heure = numbers[0] + ':' + minut
						else :
							heure = numb + ':' + minut
					else : 
						min_ar = text2number(phrase)
						print(min_ar)
						if min_ar == 0 :
								print("no")
						else :
							if numbers[0] != minut :
								heure = numbers[0] + ":" + str(min_ar)
							else :
								heure = numb + ":" + str(min_ar)
			print('yees')
			if heure.startswith(":"):
				s = heure.split(':')
				print(s[1])
				l = []
				for i in s[1] :
					l.append(i)
				if len(l)==2 :
					heure = '0' +l[1] + ':' + l[0] + '0'
			print("yeees")
			if numb != "" and heure == "" :
					minut = "00"
					for r in range (len(toks)) :
						if toks[r] == 'دقيقة' or toks[r] == "دقيقه" :
							minut = toks[r-1]
					if minut.isdigit():

						heure = numb + ":" +minut
					else :
						min_ar = text2number(phrase)
						print(min_ar)
						if min_ar == 0 :
								print("no")
						else : 
								heure = numb + ":" + str(min_ar)
#tranche d'horaire


			horaire = pd.read_csv('csv_files/horaires_oncf.csv',sep=';',encoding="utf_8") 
			horaires= np.array(horaire['horaire'])
			print("ok")
			value_horaire= np.array(horaire['value'])
			print("ok")
			ent_hor = {}
			tranche_horaire = ""
			for i in range(len(horaire)):
				ent_hor[horaires[i]] = value_horaire[i]
			for word in tokens:
				for key in ent_hor:
					if key == word:
						tranche_horaire = ent_hor[key]


			if entit_jour == "":
				if tranche_horaire != "":
					entits_jour = date[0]
					datetimeobject = dt.strptime(entits_jour ,'%Y-%m-%d')
					entit_jour  = datetimeobject.strftime('%d/%m/%Y')

						
						
			



			print("yeeees")
			if heure.startswith("3") :
				heure = heure.replace("3:","15:")
			if heure.startswith("03") :
				heure = heure.replace("03:","15:")
			if heure.startswith("2") :
				heure = heure.replace("2:","14:")
			if heure.startswith("02") :
				heure = heure.replace("02:","14:")
			if heure.startswith("4") :
				heure = heure.replace("4:","16:")
			if heure.startswith("04") :
				heure = heure.replace("04:","16:")
			if heure.startswith("5") :
				heure = heure.replace("5:","17:")
			if heure.startswith("05") :
				heure = heure.replace("05:","17:")
			if heure.startswith("6") :
				heure = heure.replace("6:","18:")
			if heure.startswith("06") :
				heure = heure.replace("06:","18:")
			if heure.startswith("01") :
				heure = heure.replace("01:","13:")
			if heure.startswith("7") :
				heure = heure.replace("7:","19:")
			if heure.startswith("07") :
				heure = heure.replace("07:","19:")

			print("yeeeeees")

#detection de classe_carte


			data_classe = pd.read_csv('csv_files/classe_carte.csv',sep=';',encoding="utf_8")
			print("ok11")
			classe_val = np.array(data_classe['classe'])
			print("ok12")
			classe_code = np.array(data_classe['value'])
			print("ok13")
			ent_cl = {}
			entity_classe = ""

			for i in range(len(classe_val)):
				ent_cl[classe_val[i]] = classe_code[i]

			print("ok1")
			match = ''
			names_word =[]
			
			for key in ent_cl:
				classe = re.search(key,msg)
				
				if classe != None :
					print("ok2")
					match_cl = classe.group()
					print("MATCH")
					
					entity_classe = str(ent_cl[key])
					print("entity_class")
					
					break
			

			if entity_classe == "":
				txt = message.text
				
				print("txt")
				for key in ent_cl:
					classe = re.search(key,txt)
				
					if classe != None :
						print("ok2")
						match_cl = classe.group()
						
						
						entity_classe = str(ent_cl[key])
						print("entity_class!!!!!!!!!!!!!!!!!!!!!!!")
						print(entity_classe)
				




			



			


#Detection des cartes d'abonnement




			data_carte = pd.read_csv('csv_files/carte_oncf.csv',sep=';',encoding="utf_8")
			print("ok11")
			carte_val = np.array(data_carte['carte'])
			print("ok12")
			carte_code = np.array(data_carte['value'])
			print("ok13")
			ent_card = {}
			entity_carte = ""

			for i in range(len(carte_val)):
				ent_card[carte_val[i]] = carte_code[i]

			print("ok1")
			match_card = ''
			names_word =[]
			
			for key in ent_card:
				card = re.search(key,msg)
				
				if card != None :
					print("ok2")
					match_card = card.group()
					print("MATCH")
					
					entity_carte = str(ent_card[key])
					print("entity_carte")
					print(entity_carte)
					break
				else :
					print("oknew00")
			print("oknew")
			code_card = ""
			if entity_classe=="1":
				if entity_carte == "NGC":
					code_card = "NGC1"
				
				if entity_carte == "NB":
					code_card = "NB1"
				
				if entity_carte == "NH":
					code_card = "NH1"

				if entity_carte == "NRS":
					code_card = "NRS1"

				if entity_carte == "FF":
					code_card = "FP"
				
				if entity_carte == "DB":
					code_card = "DB"
				
				if entity_carte == "N":
					code_card = "AC"
				
				if entity_carte == "10ZEN":
					code_card = "DA"
				if entity_carte == "DM":
					code_card = "DM"
			print("oknew2")
			if entity_classe=="2":
				if entity_carte == "NGC":
					code_card = "NGC2"
				
				if entity_carte == "NH":
					code_card = "NH2"

				if entity_carte == "NRS":
					code_card = "NRS2"

				if entity_carte == "FF":
					code_card = "FD"
				
				if entity_carte == "DB":
					code_card = "DQ"
				
				if entity_carte == "N":
					code_card = "AY"
				
				if entity_carte == "NB":
					code_card = "NB2"
				
				if entity_carte == "10ZEN":
					code_card = "DZ"
				if entity_carte == "DM":
					code_card = "MD"
			
			if entity_classe == "":
				code_card = entity_carte

			print("oknew3")

			entity_code_card = self.convert_to_code_card(code_card)

			entity_card = self.convert_to_card(entity_carte)
			entity_classe = self.convert_to_classe(entity_classe)
					

			entity_heure = self.convert_to_heure(heure)
			entity_tranche_horaire = self.convert_to_rasa_horaire(tranche_horaire)

			month = []
			month = self.convert_to_rasa_month(entit_mois)
			
			date_conv = self.convert_to_rasa_day(entit_jour)
			numbers = self.convert_to_rasa_numbers(entity_numb)
			
			


			if cities_dep == [''] and gare_dep != []:
				gare_dep_conv = self.convert_gare_dep(gare_dep)
				print("yes1")
			if gare_dep == [] and cities_dep != ['']:
				print("yes2")
				gare_dep_conv = self.convert_ville_dep(cities_dep)
			if cities_dep == [''] and gare_dep == []:
				print("yes3")
				gare_dep_conv = self.convert_gare_dep(cities_dep)
			if cities_dep != [''] and gare_dep != []:
				print("yes4")
				print(cities_dep)
				print(gare_dep)
				gare_dep_conv = self.convert_gare_dep(gare_dep)
			if cities_arr == [''] and gare_arr != []:
				print("yes5")
				gare_arr_conv = self.convert_gare_arr(gare_arr)
			if cities_arr != [''] and gare_arr ==[]:
				print("yes6")
				gare_arr_conv = self.convert_ville_arr(cities_arr)
			if cities_arr == [''] and gare_arr ==[]:
				print("yes7")
				gare_arr_conv = self.convert_gare_arr(cities_arr)
			if cities_arr != [''] and gare_arr != []:
				print("yes8")
				print(cities_arr)
				print(cities_dep)
				gare_arr_conv = self.convert_gare_arr(gare_arr)
			
			


			print("ok0")
			for t in tokens :
				print(t)
				print(len(t))
			print("ok1")
			print(gare_arr)
			print("1")
			print(cities_arr)
			print("2")
			print(gare_dep)

			print(entit_mois)
			

		

			print(gare_dep_conv)
			print(gare_arr_conv)
			print([entity_classe])
			#message.set("city_arr", [city_arr_conv], add_to_output=True)
			print("ok1")
			#message.set("city_depart", [city_dep_conv], add_to_output=True)
			print("ok1")
			
			
			message.set("gare_depart", gare_dep_conv, add_to_output=True)
			print("ok1")
			message.set("CODE_CARD", code_card, add_to_output=True)
			print("ok2")
			message.set("gare_arr", gare_arr_conv, add_to_output=True)
			print("ok2")
			message.set("date", [date_conv], add_to_output=True)
			print("ok1")
			message.set("number", numbers, add_to_output=True)
			print("ok1")
			message.set("mois", month, add_to_output=True)
			
			print("okk")
			message.set("tranche_horaire", entity_tranche_horaire, add_to_output=True)
			print("okk2")
			message.set("Classe", [entity_classe], add_to_output=True)
			print("ok1")
			message.set("Card_oncf", [entity_card], add_to_output=True)
			print("ok1")
			


from rasa.nlu.components import Component
from rasa.nlu import utils
from rasa.nlu.model import Metadata
import os
import pandas as pd
import typing
from typing import Any, Optional, Text, Dict
import numpy as np
import regex as re
from time import strftime
from datetime import datetime as dt
from typing import Any, Optional, Text, Dict
import numpy as np
import dateutil.parser as dparser
import datefinder

from dateutil.parser import parse
import datetime
import dateutil.parser as dparser
from pyarabic.number import text2number
import datefinder
from dateutil.parser import parse
from datetime import timedelta
from dateutil.relativedelta import relativedelta


from rasa.nlu.training_data import Message, TrainingData


class Extracteur_oncf(Component):
	"""A custom sentiment analysis component"""
	name = "DATA_EXTRACTION"
	provides = ["entities"]
	requires = ["tokens"]
	defaults = {}
	language_list = ["en"]
	print('initialised the class')

	def _init_(self, component_config=None):
		super(Extracteur_oncf, self)._init_(component_config)

	def train(self, training_data, cfg, **kwargs):
		"""Load the sentiment polarity labels from the text
		   file, retrieve training tokens and after formatting
		   data train the classifier."""

	
	def convert_to_entity(self, value):
		"""Convert model output into the Rasa NLU compatible output format.""" 

		entity = {"value": value,
				  
				  "entity": "ENTITY",
				  "extractor": "extractor"}

		return entity	


	def process(self, message:Message , **kwargs):
		"""Retrieve the tokens of the new message, pass it to the classifier
			and append prediction results to the message class."""
		print(message.text)	
		if not self :
			entities = []
		else:

			tokens = [t.text for t in message.get("tokens")]
			print('***********tokens*****')
			print(tokens)
            data = pd.read_csv('csv_files/fichier_entities.csv',sep=';',encoding="utf_8")
			Entity= np.array(datas['ent'])
			Value= np.array(datas['value'])
			ent_val = {}
            entities = []
			
            for i in range(len(Tville)):
				ent_val[Entity] = Value[i]
			for word in tokens:
				for key in ent_val:
					if key == word:
						print(word)
						entities.append(word)
						print(" extracted Token ++++++++++ "+ word)


            entity_ex = self.convert_to_entity(entities)
            message.set("entities_heure", [entity_ex], add_to_output=True)


import re
from typing import Any, Dict, List, Text

from rasa.nlu.tokenizers.tokenizer_fr import Token, Tokenizer
from rasa.nlu.training_data import TrainingData, Message
from typing import Text, List, Optional, Dict, Any
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.constants import TOKENS_NAMES, MESSAGE_ATTRIBUTES, TEXT,INTENT
from custom_func.translator_darijaFr_DarijaAr import trans 
from custom_func.generator import generateur_Ar
from itertools import chain
import jellyfish  #correctionPackage"
import unicodedata
from pyarabic.unshape import unshaping_line
import arabic_reshaper
from custom_func.translator_darijaFr_DarijaAr import trans 
from custom_func.generateur_cmr_ar import generateur_cmr_ar
import pyarabic.araby as araby
import arabic_reshaper
from itertools import *



class WhitespaceTokenizer_cmr_ar(Tokenizer):

    defaults = {
        # Flag to check whether to split intents
        "intent_tokenization_flag": False,
        # Symbol on which intent should be split
        "intent_split_symbol": "_",
        # Text will be tokenized with case sensitive as default
        "case_sensitive": True,
    }

    def __init__(self, component_config: Dict[Text, Any] = None) -> None:
        """Construct a new tokenizer using the WhitespaceTokenizer framework."""

        super().__init__(component_config)

        self.case_sensitive = self.component_config["case_sensitive"]
    
    

    def tokenize(self, message: Message, attribute: Text) -> List[Token]:
        text = message.get(attribute)

        if not self.case_sensitive:
            text = text.lower()

        # remove 'not a word character' if
        
        words = re.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w_~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        
        print("okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk")
        
        print("****************************************************************************")
        print(words)
        print("****************************************************************************")
        
        
        
        tokens = []
        running_offset = 0
        
        stop_words = ['واحد','ن','لس','ف','هاد','نسولك','عافاك','ممكن','مومكين','في','ف','شي','تقدر','بغيت', 'بغهيت','ديال','ديالكم','واش','ل','نقدر','لس','لك','لو','س']
        for word in words:
            if word in stop_words:
                continue
            else :

                word_offset = text.index(word, running_offset)
                word_len = len(word)
                running_offset = word_offset + word_len
                tokens.append(Token(word, word_offset))


        return tokens
        
    def process(self, message: Message, **kwargs: Any) -> None:
        """Tokenize the incoming message."""

        lettre="abcdefghijklmnopqrstuvwxyzABSCDEFGHIJKLMNOPQRSTUVWXYZ"
        nbre = "123546879"
        arabe="ابتثحخدذرزسشصضطظعغفقكمنهويءإأٱآةؤئى"
        print("heeeeeeeeeeeere")
        txt = message.text
        txt.lower()
        print(txt)
        nb = 0
        for i in txt:
            print(i+"process")
            if i in arabe:
               
                tokens = self.tokenize_arab(message, TEXT)
                nb = 1
                break
            elif i in nbre :
                continue
            else :
                tokens = self.tokenize_dr(message, TEXT)
                nb =1 
                break
        if nb == 0 :
            tokens = self.tokenize_arab(message, TEXT)
        
        tokens = self.add_cls_token(tokens, TEXT)
        message.set(TOKENS_NAMES[TEXT], tokens)
    
    def train(
        self,
        training_data: TrainingData,
        config: Optional[RasaNLUModelConfig] = None,
        **kwargs: Any,
    ) -> None:
        generateur_cmr_ar('data_cmr_ar')
        """Tokenize all training data.""" 

        for example in training_data.training_examples:
            for attribute in MESSAGE_ATTRIBUTES:
                if example.get(attribute) is not None:
                    if attribute == INTENT:
                        tokens = self._split_intent(example)
                    else:
                        tokens = self.tokenize(example, attribute)
                        tokens = self.add_cls_token(tokens, attribute)
                    example.set(TOKENS_NAMES[attribute], tokens)


    
    def tokenize_dr(self, message: Message, attribute: Text ) -> List[Token]:
        print("darijaaaaaaaaaa")
        text = message.get(attribute)

        
        text = text.lower()
        numb = []

        toks = ''
        toks = text.split()
        print("tokssssssssssssssssss")
        numbs = []
        for t in toks:
            
            w=re.search(r'((([\d|\d\.\d]+)+(kg|kgs|g|kilo|كيلو|كلغ|كغ|غ|ج|غرام)))', t)
            if w != None :
                numbs.append(w.group())
                
                print(w[0])
                print('group+++++++++++++++++++++')
             
        
        exp = ['kg','kgs','g','kilo','kilos']
        for word in toks:
            if word in exp :
                print("im here1")
                if toks[toks.index(word)-1].isdigit():
                    print("im here2")
                    numbs.append(word)
                    numbs.append(toks[toks.index(word)-1])


        for s in toks:
            if s.isdigit():
               numb.append(s)
        
        text = trans(text)
        # print(numb)
        print("texttranslateld******************"+ text)
        for i in range(0,len(numb)):

            text = text+" "+ str(numb[i]) + " " 
        
        for i in range(0,len(numbs)):

            text = text+" "+ str(numbs[i]) + " " 
        
        # remove 'not a word character' if
        print(text)

        
        words = re.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w_~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        
        print("okkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkkk")
        with open("dictionaries/dict_cmr_ar.txt", 'r',encoding="utf8") as f:
            list = set(chain(*(line.split() for line in f if line)))
        print("****************************************************************************")
        print(words)
        print("****************************************************************************")
        dicto = {}
        # if we removed everything like smiles `:)`, use the whole text as 1 token
        #words2 = []
        
        tokens = []
        running_offset = 0
        print("////////////////////////////////////////////////////////")
        
        stop_words = ['واحد','ن','لس','ف','هاد','نسولك','عافاك','ممكن','مومكين','في','ف','شي','تقدر','بغيت', 'بغهيت','ديال','ديالكم','واش','ل','نقدر','لس','لك','لو','س']
        for word in words:
            w=re.search(r'((([\d|\d\.\d]+)+(kg|kgs|g|kilo|كغ|كيلو|كلغ|غ|ج|غرام)))', word)
            
            
            if word.isnumeric() :
                min_dist = word
            else: 
                if w != None :
                    print("yes weight")
                    min_dist = w.group()
                else:
                    for mot in list:
                        dicto[mot]=jellyfish.levenshtein_distance(mot, word)
                    min_dist = min(dicto.keys(), key=(lambda k: dicto[k]))
            
            if jellyfish.levenshtein_distance(word,min_dist) > 2:
                min_dist = word
            print("min"+min_dist)
            
            if min_dist in stop_words:
                continue
            else :
                print("TOKEN"+min_dist)
                word_corr = min_dist
                word_offset = text.index(word, running_offset)
                word_len = len(word)
                running_offset = word_offset + word_len
                print("word_corr"+word_corr)
                tokens.append(Token(word_corr, word_offset))


        return tokens
   
    

    def tokenize_arab(self, message: Message, attribute: Text ) -> List[Token]:
        text = message.get(attribute)
        text.lower()

        print("araaaaaaaaaaaabe")
        words = re.sub(
            # there is a space or an end of a string after it
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        with open("dictionaries/dict_cmr_ar.txt", 'r',encoding="utf8") as f:
            list = set(chain(*(line.split() for line in f if line)))
        

        dicto = {}
        
        
        tokens = []
        running_offset = 0
        stop_words = ['ف','هاد','نسولك','عافاك','ممكن','مومكين','في','ف','شي','تقدر','بغيت', 'بغهيت','ديال','ديالكم','واش','واحد','نقدر','لس','لك','لو','س']
        for word in words:
            w=re.search(r'((([\d|\d\.\d]+)+(kg|kgs|g|kilo|كيلو|كلغ|كغ|غ|ج|غرام)))', word)
            
            
            if word.isnumeric() :
                min_dist = word
            else: 
                if w != None :
                    print("yes weight")
                    min_dist = w.group()
                else:
                    for mot in list:
                        dicto[mot]=jellyfish.levenshtein_distance(mot, word)
                    min_dist = min(dicto.keys(), key=(lambda k: dicto[k]))
            
            if jellyfish.levenshtein_distance(word,min_dist) > 2:
                min_dist = word
            print("min"+min_dist)
            
            if min_dist in stop_words:
                word_corr = ''
                word_offset = text.index(word, running_offset)
                word_len = len(word)
                running_offset = word_offset + word_len
                tokens.append(Token(word_corr, word_offset))
            else :
                word_corr = min_dist
                word_offset = text.index(word, running_offset)
                word_len = len(word)
                running_offset = word_offset + word_len
                tokens.append(Token(word_corr, word_offset))
            


        return tokens


"""This is a somewhat delicate package. It contains all registered components
and preconfigured templates.

Hence, it imports all of the components. To avoid cycles, no component should
import this in module scope."""

import logging
import typing
from typing import Any, Dict, List, Optional, Text, Type

from rasa.constants import DOCS_URL_COMPONENTS

from rasa.nlu.classifiers.diet_classifier import DIETClassifier
from rasa.nlu.classifiers.keyword_intent_classifier import KeywordIntentClassifier
from rasa.nlu.classifiers.mitie_intent_classifier import MitieIntentClassifier
from rasa.nlu.classifiers.sklearn_intent_classifier import SklearnIntentClassifier
from rasa.nlu.classifiers.embedding_intent_classifier import EmbeddingIntentClassifier
from rasa.nlu.extractors.crf_entity_extractor import CRFEntityExtractor
from rasa.nlu.extractors.duckling_http_extractor import DucklingHTTPExtractor
from rasa.nlu.extractors.entity_synonyms import EntitySynonymMapper
from rasa.nlu.extractors.mitie_entity_extractor import MitieEntityExtractor
from rasa.nlu.extractors.spacy_entity_extractor import SpacyEntityExtractor
from rasa.nlu.extractors.extract_dupont import Extract_dupont
from rasa.nlu.featurizers.sparse_featurizer.lexical_syntactic_featurizer import (
    LexicalSyntacticFeaturizer,
)
from rasa.nlu.featurizers.dense_featurizer.convert_featurizer import ConveRTFeaturizer
from rasa.nlu.featurizers.dense_featurizer.mitie_featurizer import MitieFeaturizer
from rasa.nlu.featurizers.dense_featurizer.spacy_featurizer import SpacyFeaturizer
from rasa.nlu.featurizers.sparse_featurizer.count_vectors_featurizer import (
    CountVectorsFeaturizer,
)
from rasa.nlu.featurizers.dense_featurizer.lm_featurizer import LanguageModelFeaturizer
from rasa.nlu.featurizers.sparse_featurizer.regex_featurizer import RegexFeaturizer
from rasa.nlu.model import Metadata
from rasa.nlu.selectors.response_selector import ResponseSelector
from rasa.nlu.tokenizers.convert_tokenizer import ConveRTTokenizer
from rasa.nlu.tokenizers.jieba_tokenizer import JiebaTokenizer
from rasa.nlu.tokenizers.mitie_tokenizer import MitieTokenizer
from rasa.nlu.tokenizers.spacy_tokenizer import SpacyTokenizer
from rasa.nlu.tokenizers.whitespace_tokenizer import WhitespaceTokenizer
from rasa.nlu.tokenizers.whitespace2_omran import WhitespaceTokenizer2
from rasa.nlu.tokenizers.whitespace_ch import Whitespace_ch
from rasa.nlu.tokenizers.whitespacetokenizer_arab import WhitespaceTokenizer_arab
from rasa.nlu.tokenizers.whitespacetokenizer_oncf_ar import WhitespaceTokenizer_oncf_ar
from rasa.nlu.tokenizers.WhitespaceTokenizer_oncf_ar_v2 import WhitespaceTokenizer_oncf_ar_v2
from rasa.nlu.tokenizers.WhitespaceTokenizer_ecom_ar import WhitespaceTokenizer_ecom_ar
from rasa.nlu.tokenizers.WhitespaceTokenizer_ecom_fr import WhitespaceTokenizer_ecom_fr
from rasa.nlu.tokenizers.WhitespaceTokenizer_ecom_eng import WhitespaceTokenizer_ecom_ang
from rasa.nlu.tokenizers.lm_tokenizer import LanguageModelTokenizer
from rasa.nlu.utils.mitie_utils import MitieNLP
from rasa.nlu.utils.spacy_utils import SpacyNLP
from rasa.nlu.utils.hugging_face.hf_transformers import HFTransformersNLP
from rasa.utils.common import class_from_module_path, raise_warning
from rasa.nlu.tokenizers.whitespacetokenizer_omran_fr import WhitespaceTokenizer_omran_fr
from rasa.nlu.tokenizers.whitespacetokenizer_oncf_fr import WhitespaceTokenizer_oncf_fr
from rasa.nlu.tokenizers.whitespacetokenizer_oncf_fr_v2 import WhitespaceTokenizer_oncf_fr_v2
from rasa.nlu.tokenizers.whitespacetokenizer_omran_ar import WhitespaceTokenizer_omran_ar
from rasa.nlu.tokenizers.Whitespacetokenizer_omrfr import WhitespaceTokenizer_omr_fr
from rasa.nlu.tokenizers.whitespacetokenizer_ar_cdg import WhitespaceTokenizer_ar_cdg

from rasa.nlu.tokenizers.WhitespaceTokenizer_oar import WhitespaceTokenizer_oar

from rasa.nlu.extractors.extracteur_cdg import Extracteur_cdg 
from rasa.nlu.extractors.Extracteur_oncf import Extracteur_oncf
from rasa.nlu.extractors.Extracteur_place_time import Extracteur_place_time

from rasa.nlu.extractors.Omran_extracteur  import omran_Extracteur
from rasa.nlu.extractors.Extractor_ecommerce import Extractor_ecommerce
from rasa.nlu.extractors.Extractor_ecommerce_ar import Extractor_ecommerce_ar
from rasa.nlu.tokenizers.tokenizer_ecomm_test_ar import WhitespaceTokenizer_ecomm_info_ar
from rasa.utils.tensorflow.constants import (
    INTENT_CLASSIFICATION,
    ENTITY_RECOGNITION,
    NUM_TRANSFORMER_LAYERS,
)
from rasa.nlu.extractors.extracteur_chrono import Extracteur_chrono
from rasa.nlu.extractors.extract_chrono import Extract_chrono
from rasa.nlu.extractors.omran_produit_extractor  import Extracteur_omran
if typing.TYPE_CHECKING:
    from rasa.nlu.components import Component
    from rasa.nlu.config import RasaNLUModelConfig, RasaNLUModelConfig

logger = logging.getLogger(__name__)


# Classes of all known components. If a new component should be added,
# its class name should be listed here.
component_classes = [
    # utils
    SpacyNLP,
    MitieNLP,
    HFTransformersNLP,
    # tokenizers
    WhitespaceTokenizer_ar_cdg,
    MitieTokenizer,
    SpacyTokenizer,
    WhitespaceTokenizer,
    WhitespaceTokenizer2,
    WhitespaceTokenizer_ecom_ar,
    WhitespaceTokenizer_ecom_fr,
    WhitespaceTokenizer_ecom_ang,
    WhitespaceTokenizer_oncf_fr,
    ConveRTTokenizer,
    JiebaTokenizer,
    LanguageModelTokenizer,
    WhitespaceTokenizer_arab,
    WhitespaceTokenizer_omran_ar,
    WhitespaceTokenizer_omran_fr,
    WhitespaceTokenizer_oar,
    WhitespaceTokenizer_oncf_ar_v2,
    WhitespaceTokenizer_oncf_ar,
    WhitespaceTokenizer_omr_fr,
    WhitespaceTokenizer_ecomm_info_ar,
    WhitespaceTokenizer_oncf_fr_v2,
    Whitespace_ch,
    # extractors
    Extracteur_cdg,
    Extracteur_omran,
    Extractor_ecommerce_ar,
    Extracteur_chrono,
    Extract_chrono,
    Extracteur_oncf,
    SpacyEntityExtractor,
    MitieEntityExtractor,
    CRFEntityExtractor,
    DucklingHTTPExtractor,
    EntitySynonymMapper,
    omran_Extracteur,
    Extracteur_place_time,
    Extractor_ecommerce,
    # featurizers
    SpacyFeaturizer,
    MitieFeaturizer,
    RegexFeaturizer,
    LexicalSyntacticFeaturizer,
    CountVectorsFeaturizer,
    ConveRTFeaturizer,
    LanguageModelFeaturizer,
    # classifiers
    SklearnIntentClassifier,
    MitieIntentClassifier,
    KeywordIntentClassifier,
    DIETClassifier,
    EmbeddingIntentClassifier,
    # selectors
    ResponseSelector,
]

# Mapping from a components name to its class to allow name based lookup.
registered_components = {c.name: c for c in component_classes}

# DEPRECATED ensures compatibility, will be remove in future versions
old_style_names = {
    "nlp_spacy": "SpacyNLP",
    "nlp_mitie": "MitieNLP",
    "ner_spacy": "SpacyEntityExtractor",
    "ner_mitie": "MitieEntityExtractor",
    "ner_crf": "CRFEntityExtractor",
    "ner_duckling_http": "DucklingHTTPExtractor",
    "ner_synonyms": "EntitySynonymMapper",
    "intent_featurizer_spacy": "SpacyFeaturizer",
    "intent_featurizer_mitie": "MitieFeaturizer",
    "intent_featurizer_ngrams": "NGramFeaturizer",
    "intent_entity_featurizer_regex": "RegexFeaturizer",
    "intent_featurizer_count_vectors": "CountVectorsFeaturizer",
    "tokenizer_mitie": "MitieTokenizer",
    "tokenizer_spacy": "SpacyTokenizer",
    "tokenizer_whitespace": "WhitespaceTokenizer",
    "tokenizer_jieba": "JiebaTokenizer",
    "intent_classifier_sklearn": "SklearnIntentClassifier",
    "intent_classifier_mitie": "MitieIntentClassifier",
    "intent_classifier_keyword": "KeywordIntentClassifier",
    "intent_classifier_tensorflow_embedding": "EmbeddingIntentClassifier",
}

# To simplify usage, there are a couple of model templates, that already add
# necessary components in the right order. They also implement
# the preexisting `backends`.
registered_pipeline_templates = {
    "pretrained_embeddings_spacy": [
        {"name": "SpacyNLP"},
        {"name": "SpacyTokenizer"},
        {"name": "SpacyFeaturizer"},
        {"name": "RegexFeaturizer"},
        {"name": "CRFEntityExtractor"},
        {"name": "EntitySynonymMapper"},
        {"name": "SklearnIntentClassifier"},
    ],
    "keyword": [{"name": "KeywordIntentClassifier"}],
    "supervised_embeddings": [
        {"name": "WhitespaceTokenizer"},
        {"name": "RegexFeaturizer"},
        {"name": "CRFEntityExtractor"},
        {"name": "EntitySynonymMapper"},
        {"name": "CountVectorsFeaturizer"},
        {
            "name": "CountVectorsFeaturizer",
            "analyzer": "char_wb",
            "min_ngram": 1,
            "max_ngram": 4,
        },
        {"name": "EmbeddingIntentClassifier"},
    ],
    "pretrained_embeddings_convert": [
        {"name": "ConveRTTokenizer"},
        {"name": "ConveRTFeaturizer"},
        {"name": "EmbeddingIntentClassifier"},
    ],
}


def pipeline_template(s: Text) -> Optional[List[Dict[Text, Any]]]:
    import copy

    # do a deepcopy to avoid changing the template configurations
    return copy.deepcopy(registered_pipeline_templates.get(s))


def get_component_class(component_name: Text) -> Type["Component"]:
    """Resolve component name to a registered components class."""

    if component_name not in registered_components:
        if component_name not in old_style_names:
            try:
                return class_from_module_path(component_name)

            except AttributeError:
                # when component_name is a path to a class but the path does not contain
                # that class
                module_name, _, class_name = component_name.rpartition(".")
                raise Exception(
                    f"Failed to find class '{class_name}' in module '{module_name}'.\n"
                )
            except ImportError as e:
                # when component_name is a path to a class but that path is invalid or
                # when component_name is a class name and not part of old_style_names

                is_path = "." in component_name

                if is_path:
                    module_name, _, _ = component_name.rpartition(".")
                    exception_message = f"Failed to find module '{module_name}'. \n{e}"
                else:
                    exception_message = (
                        f"Cannot find class '{component_name}' from global namespace. "
                        f"Please check that there is no typo in the class "
                        f"name and that you have imported the class into the global "
                        f"namespace."
                    )

                raise ModuleNotFoundError(exception_message)
        else:
            # DEPRECATED ensures compatibility, remove in future versions
            raise_warning(
                f"Your nlu config file "
                f"contains old style component name `{component_name}`, "
                f"you should change it to its new class name: "
                f"`{old_style_names[component_name]}`.",
                FutureWarning,
                docs=DOCS_URL_COMPONENTS,
            )
            component_name = old_style_names[component_name]

    return registered_components[component_name]


def load_component_by_meta(
    component_meta: Dict[Text, Any],
    model_dir: Text,
    metadata: Metadata,
    cached_component: Optional["Component"],
    **kwargs: Any,
) -> Optional["Component"]:
    """Resolves a component and calls its load method.

    Inits it based on a previously persisted model.
    """

    # try to get class name first, else create by name
    component_name = component_meta.get("class", component_meta["name"])
    component_class = get_component_class(component_name)
    return component_class.load(
        component_meta, model_dir, metadata, cached_component, **kwargs
    )


def create_component_by_config(
    component_config: Dict[Text, Any], config: "RasaNLUModelConfig"
) -> Optional["Component"]:
    """Resolves a component and calls it's create method.

    Inits it based on a previously persisted model.
    """

    # try to get class name first, else create by name
    component_name = component_config.get("class", component_config["name"])
    component_class = get_component_class(component_name)
    return component_class.create(component_config, config)



from typing import Any, Dict, List, Text

from rasa.nlu.tokenizers.tokenizer import Token
from rasa.nlu.tokenizers.whitespace_tokenizer import WhitespaceTokenizer
from rasa.nlu.training_data import Message
from rasa.nlu.constants import MESSAGE_ATTRIBUTES, TOKENS_NAMES
import rasa.utils.train_utils as train_utils
import tensorflow as tf


class ConveRTTokenizer(WhitespaceTokenizer):
    """Tokenizer using ConveRT model.

    Loads the ConveRT(https://github.com/PolyAI-LDN/polyai-models#convert)
    model from TFHub and computes sub-word tokens for dense
    featurizable attributes of each message object.
    """

    defaults = {
        # Flag to check whether to split intents
        "intent_tokenization_flag": False,
        # Symbol on which intent should be split
        "intent_split_symbol": "_",
        # Text will be tokenized with case sensitive as default
        "case_sensitive": True,
    }

    def __init__(self, component_config: Dict[Text, Any] = None) -> None:
        """Construct a new tokenizer using the WhitespaceTokenizer framework."""

        super().__init__(component_config)

        model_url = "http://models.poly-ai.com/convert/v1/model.tar.gz"
        self.module = train_utils.load_tf_hub_model(model_url)

        self.tokenize_signature = self.module.signatures["tokenize"]

    def _tokenize(self, sentence: Text) -> Any:

        return self.tokenize_signature(tf.convert_to_tensor([sentence]))[
            "default"
        ].numpy()

    def tokenize(self, message: Message, attribute: Text) -> List[Token]:
        """Tokenize the text using the ConveRT model.

        ConveRT adds a special char in front of (some) words and splits words into
        sub-words. To ensure the entity start and end values matches the token values,
        tokenize the text first using the whitespace tokenizer. If individual tokens
        are split up into multiple tokens, we make sure that the start end end value
        of the first and last respective tokens stay the same.
        """

        # perform whitespace tokenization
        tokens_in = super().tokenize(message, attribute)

        tokens_out = []

        for token in tokens_in:
            token_start, token_end, token_text = token.start, token.end, token.text

            # use ConveRT model to tokenize the text
            split_token_strings = self._tokenize(token_text)[0]

            # clean tokens (remove special chars and empty tokens)
            split_token_strings = self._clean_tokens(split_token_strings)

            tokens_out += train_utils.align_tokens(
                split_token_strings, token_end, token_start
            )

        return tokens_out

    def _clean_tokens(self, tokens: List[bytes]):
        """Encode tokens and remove special char added by ConveRT."""

        tokens = [string.decode("utf-8").replace("﹏", "") for string in tokens]
        return [string for string in tokens if string]



from rasa.nlu.components import Component
from rasa.nlu import utils
from rasa.nlu.model import Metadata
import os
import pandas as pd
import typing
from typing import Any, Optional, Text, Dict
import numpy as np
import regex as re


class Extract_chrono(Component):
	"""A custom sentiment analysis component"""
	name = "FACT_EXTRACTION"
	provides = ["entities"]
	requires = ["tokens"]
	defaults = {}
	language_list = ["en"]
	print('initialised the class')

	def __init__(self, component_config=None):
		super(Extract_chrono, self).__init__(component_config)

	def train(self, training_data, cfg, **kwargs):
		"""Load the sentiment polarity labels from the text
		   file, retrieve training tokens and after formatting
		   data train the classifier."""



	def convert_to_rasa_pays(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "PAYS",
				"extractor": "extractor"}

		return entity

	def convert_to_rasa_poids(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "POIDS",
				"extractor": "extractor"}

		return entity

	def convert_to_rasa_product(self, value):
		"""Convert model output into the Rasa NLU compatible output format."""

		
		entity = {"value": value,
				  
				"entity": "VILLE",
				"extractor": "extractor"}

		return entity
		


	def process(self, message, **kwargs):
		"""Retrieve the tokens of the new message, pass it to the classifier
			and append prediction results to the message class."""
		if not self :
			entities = []
		else:
	
#detection pays	
			print("ok")
			tokens = [t.text for t in message.get("tokens")]
			datas = pd.read_csv('countries_ar.csv',sep=';',encoding="utf_8")
			code= np.array(datas['code'])
			pays= np.array(datas['name'])
			ent_val = {}
			entity_conv = []
			entity = ""
			for i in range(len(code)):
				ent_val[pays[i]] = code[i]
			print("ok2")
			for word in tokens:
				print(word)
				for key in ent_val:
					if key == word:
						print("pays trouvé !!")
						entity = ent_val[key]
						
			entity_conv = self.convert_to_rasa_pays(entity)

#detection de poids !!!!
			number =""
			weight =[]	
			toks = ""
			numb = ""
			gram = ['g','gram','gr','ج','غرام','غ','كرام','جرام']





			for t in tokens :
				toks += t  
				print(t+" token")
				w=re.search(r'((([\d|\d\.\d]+)+(kg|kilo|كج|كلغ|كغ|كيلو|كيلوغرام)))', t)
				if w == None :
					print("no kg weight +++++++++++++++++++") 
				else :
					weight = w.string
					print(weight+"+++++++++++++++weight(kg)")
					for i in weight :
						print("digiiiiiiiit "+ i)
						if i.isdigit() or i == ".":
							print("digiiiiiiiit")
							number += i
							print(i)
						else:
							break
					numb += number
					print(numb+"number")

			if numb == "" :
				for t in tokens :
					toks += t  
					print(t+"token")
					w=re.search(r'((([\d|\d\.\d]+)+(gr|gram|g|غ|ج|غرام)))', t)
					if w == None :
						print("no g weight")
					else :
						weight = w.string
						print(weight+"weight(g)")
						for i in weight :
							if i.isdigit():
								number += i
							else:
								break
						numb += str(int(number)*(0.001))
			
			if numb == "":
				print("no weight regex")
				kilo = ['kg','kgs','kilos','kilo','كج','كلغ','كيلو','كغ','kilogramme','kilogram']
                
				for word in tokens:
					print("word"+word)
					if word in kilo :
						print("there is a weight(kg)")
						if tokens[tokens.index(word)-1].isdigit() :
							print("l9ina weight(kg)")
							numb = tokens[tokens.index(word)-1]
						if tokens[tokens.index(word)+1].isdigit() :
							print("l9ina weight(kg)")
							numb = tokens[tokens.index(word)-1]

				for word in tokens:
					print("word"+word)
					if word in gram :
						print("there is a weight(gram)")
						if tokens[tokens.index(word)-1].isdigit() :
							print("tokens[tokens.index(word)-1])"+tokens[tokens.index(word)-1])
							numb = str(int(tokens[tokens.index(word)-1])*(0.001))
						if tokens[tokens.index(word)+1].isdigit() :
							print("tokens[tokens.index(word)+1]"+tokens[tokens.index(word)+1])
							numb =str(int(tokens[tokens.index(word)+1])*(0.001))
                
			else :
				print("weight regex ok")

#detection produit
			data_prod = pd.read_csv('csv_ville.csv',sep=';',encoding="utf_8")

			code_prod= np.array(data_prod['ville'])
			prod= np.array(data_prod['value'])
			ent_val2 = {}
			entity_conv2 = []
			entity2 = ""
			for i in range(len(code_prod)):
				ent_val2[prod[i]] = code_prod[i]
			for w in tokens:
				for key2 in ent_val2:
					if key2 == w:
						entity2 = ent_val2[key2]


			entity_conv2 = self.convert_to_rasa_product(entity2)			
			entity_conv = self.convert_to_rasa_pays(entity)
			entity_poids = self.convert_to_rasa_poids(numb)


			message.set("entity_pays", [entity_conv], add_to_output=True)
			message.set("entity_ville", [entity_conv2], add_to_output=True)
			message.set("entity_poids", [entity_poids], add_to_output=True)


import asyncio

import datetime
import pytest
import time
import uuid
import json
from _pytest.monkeypatch import MonkeyPatch
from aioresponses import aioresponses
from typing import Optional, Text, List
from unittest.mock import patch

from rasa.core import jobs
from rasa.core.actions.action import ACTION_LISTEN_NAME, ACTION_SESSION_START_NAME

from rasa.core.agent import Agent
from rasa.core.channels.channel import CollectingOutputChannel, UserMessage
from rasa.core.domain import SessionConfig
from rasa.core.events import (
    ActionExecuted,
    BotUttered,
    ReminderCancelled,
    ReminderScheduled,
    Restarted,
    UserUttered,
    SessionStarted,
    Event,
    SlotSet,
)
from rasa.core.interpreter import RasaNLUHttpInterpreter
from rasa.core.processor import MessageProcessor, DEFAULT_INTENTS
from rasa.core.slots import Slot
from rasa.core.trackers import DialogueStateTracker
from rasa.utils.endpoints import EndpointConfig
from tests.utilities import latest_request

from rasa.core.constants import EXTERNAL_MESSAGE_PREFIX, IS_EXTERNAL

import logging

logger = logging.getLogger(__name__)


async def test_message_processor(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    await default_processor.handle_message(
        UserMessage('/greet{"name":"Core"}', default_channel)
    )
    assert default_channel.latest_output() == {
        "recipient_id": "default",
        "text": "hey there Core!",
    }


async def test_message_id_logging(default_processor: MessageProcessor):
    from rasa.core.trackers import DialogueStateTracker

    message = UserMessage("If Meg was an egg would she still have a leg?")
    tracker = DialogueStateTracker("1", [])
    await default_processor._handle_message_with_tracker(message, tracker)
    logged_event = tracker.events[-1]

    assert logged_event.message_id == message.message_id
    assert logged_event.message_id is not None


async def test_parsing(default_processor: MessageProcessor):
    message = UserMessage('/greet{"name": "boy"}')
    parsed = await default_processor._parse_message(message)
    assert parsed["intent"]["name"] == "greet"
    assert parsed["entities"][0]["entity"] == "name"


async def test_log_unseen_feature(default_processor: MessageProcessor):
    message = UserMessage('/dislike{"test_entity": "RASA"}')
    parsed = await default_processor._parse_message(message)
    with pytest.warns(UserWarning) as record:
        default_processor._log_unseen_features(parsed)
    assert len(record) == 2

    assert (
        record[0].message.args[0].startswith("Interpreter parsed an intent 'dislike'")
    )
    assert (
        record[1]
        .message.args[0]
        .startswith("Interpreter parsed an entity 'test_entity'")
    )


@pytest.mark.parametrize("default_intent", DEFAULT_INTENTS)
async def test_default_intent_recognized(
    default_processor: MessageProcessor, default_intent: Text
):
    message = UserMessage(default_intent)
    parsed = await default_processor._parse_message(message)
    with pytest.warns(None) as record:
        default_processor._log_unseen_features(parsed)
    assert len(record) == 0


async def test_http_parsing():
    message = UserMessage("lunch?")

    endpoint = EndpointConfig("https://interpreter.com")
    with aioresponses() as mocked:
        mocked.post("https://interpreter.com/model/parse", repeat=True, status=200)

        inter = RasaNLUHttpInterpreter(endpoint_config=endpoint)
        try:
            await MessageProcessor(inter, None, None, None, None)._parse_message(
                message
            )
        except KeyError:
            pass  # logger looks for intent and entities, so we except

        r = latest_request(mocked, "POST", "https://interpreter.com/model/parse")

        assert r


async def mocked_parse(self, text, message_id=None, tracker=None):
    """Mock parsing a text message and augment it with the slot
    value from the tracker's state."""

    return {
        "intent": {"name": "", "confidence": 0.0},
        "entities": [],
        "text": text,
        "requested_language": tracker.get_slot("requested_language"),
    }


async def test_parsing_with_tracker():
    tracker = DialogueStateTracker.from_dict("1", [], [Slot("requested_language")])

    # we'll expect this value 'en' to be part of the result from the interpreter
    tracker._set_slot("requested_language", "en")

    endpoint = EndpointConfig("https://interpreter.com")
    with aioresponses() as mocked:
        mocked.post("https://interpreter.com/parse", repeat=True, status=200)

        # mock the parse function with the one defined for this test
        with patch.object(RasaNLUHttpInterpreter, "parse", mocked_parse):
            interpreter = RasaNLUHttpInterpreter(endpoint_config=endpoint)
            agent = Agent(None, None, interpreter)
            result = await agent.parse_message_using_nlu_interpreter("lunch?", tracker)

            assert result["requested_language"] == "en"


async def test_reminder_scheduled(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = uuid.uuid4().hex

    reminder = ReminderScheduled("remind", datetime.datetime.now())
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    tracker.update(UserUttered("test"))
    tracker.update(ActionExecuted("action_schedule_reminder"))
    tracker.update(reminder)

    default_processor.tracker_store.save(tracker)

    await default_processor.handle_reminder(
        reminder, sender_id, default_channel, default_processor.nlg
    )

    # retrieve the updated tracker
    t = default_processor.tracker_store.retrieve(sender_id)

    assert t.events[-5] == UserUttered("test")
    assert t.events[-4] == ActionExecuted("action_schedule_reminder")
    assert isinstance(t.events[-3], ReminderScheduled)
    assert t.events[-2] == UserUttered(
        f"{EXTERNAL_MESSAGE_PREFIX}remind", intent={"name": "remind", IS_EXTERNAL: True}
    )
    assert t.events[-1] == ActionExecuted("action_listen")


async def test_reminder_aborted(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = uuid.uuid4().hex

    reminder = ReminderScheduled(
        "utter_greet", datetime.datetime.now(), kill_on_user_message=True
    )
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    tracker.update(reminder)
    tracker.update(UserUttered("test"))  # cancels the reminder

    default_processor.tracker_store.save(tracker)
    await default_processor.handle_reminder(
        reminder, sender_id, default_channel, default_processor.nlg
    )

    # retrieve the updated tracker
    t = default_processor.tracker_store.retrieve(sender_id)
    assert len(t.events) == 3  # nothing should have been executed


async def wait_until_all_jobs_were_executed(
    timeout_after_seconds: Optional[float] = None,
) -> None:
    total_seconds = 0.0
    while len((await jobs.scheduler()).get_jobs()) > 0 and (
        not timeout_after_seconds or total_seconds < timeout_after_seconds
    ):
        await asyncio.sleep(0.1)
        total_seconds += 0.1

    if total_seconds >= timeout_after_seconds:
        jobs.kill_scheduler()
        raise TimeoutError


async def test_reminder_cancelled_multi_user(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_ids = [uuid.uuid4().hex, uuid.uuid4().hex]
    trackers = []
    for sender_id in sender_ids:
        tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

        tracker.update(UserUttered("test"))
        tracker.update(ActionExecuted("action_reminder_reminder"))
        tracker.update(
            ReminderScheduled(
                "greet", datetime.datetime.now(), kill_on_user_message=True
            )
        )
        trackers.append(tracker)

    # cancel all reminders (one) for the first user
    trackers[0].update(ReminderCancelled())

    for tracker in trackers:
        default_processor.tracker_store.save(tracker)
        await default_processor._schedule_reminders(
            tracker.events, tracker, default_channel, default_processor.nlg
        )
    # check that the jobs were added
    assert len((await jobs.scheduler()).get_jobs()) == 2

    for tracker in trackers:
        await default_processor._cancel_reminders(tracker.events, tracker)
    # check that only one job was removed
    assert len((await jobs.scheduler()).get_jobs()) == 1

    # execute the jobs
    await wait_until_all_jobs_were_executed(timeout_after_seconds=5.0)

    tracker_0 = default_processor.tracker_store.retrieve(sender_ids[0])
    # there should be no utter_greet action
    assert (
        UserUttered(
            f"{EXTERNAL_MESSAGE_PREFIX}greet",
            intent={"name": "greet", IS_EXTERNAL: True},
        )
        not in tracker_0.events
    )

    tracker_1 = default_processor.tracker_store.retrieve(sender_ids[1])
    # there should be utter_greet action
    assert (
        UserUttered(
            f"{EXTERNAL_MESSAGE_PREFIX}greet",
            intent={"name": "greet", IS_EXTERNAL: True},
        )
        in tracker_1.events
    )


async def test_reminder_cancelled_cancels_job_with_name(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = "][]][xy,,=+2f'[:/;>]  <0d]A[e_,02"

    reminder = ReminderScheduled(
        intent="greet", trigger_date_time=datetime.datetime.now()
    )
    job_name = reminder.scheduled_job_name(sender_id)
    reminder_cancelled = ReminderCancelled()

    assert reminder_cancelled.cancels_job_with_name(job_name, sender_id)
    assert not reminder_cancelled.cancels_job_with_name(job_name.upper(), sender_id)


async def test_reminder_cancelled_cancels_job_with_name_special_name(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = "][]][xy,,=+2f'[:/;  >]<0d]A[e_,02"
    name = "wkjbgr,34(,*&%^^&*(OP#LKMN V#NF# # #R"

    reminder = ReminderScheduled(
        intent="greet", trigger_date_time=datetime.datetime.now(), name=name
    )
    job_name = reminder.scheduled_job_name(sender_id)
    reminder_cancelled = ReminderCancelled(name)

    assert reminder_cancelled.cancels_job_with_name(job_name, sender_id)
    assert not reminder_cancelled.cancels_job_with_name(job_name.upper(), sender_id)


async def cancel_reminder_and_check(
    tracker: DialogueStateTracker,
    default_processor: MessageProcessor,
    reminder_canceled_event: ReminderCancelled,
    num_jobs_before: int,
    num_jobs_after: int,
) -> None:
    # cancel the sixth reminder
    tracker.update(reminder_canceled_event)

    # check that the jobs were added
    assert len((await jobs.scheduler()).get_jobs()) == num_jobs_before

    await default_processor._cancel_reminders(tracker.events, tracker)

    # check that only one job was removed
    assert len((await jobs.scheduler()).get_jobs()) == num_jobs_after


async def test_reminder_cancelled_by_name(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    tracker_with_six_scheduled_reminders: DialogueStateTracker,
):
    tracker = tracker_with_six_scheduled_reminders
    await default_processor._schedule_reminders(
        tracker.events, tracker, default_channel, default_processor.nlg
    )

    # cancel the sixth reminder
    await cancel_reminder_and_check(
        tracker, default_processor, ReminderCancelled("special"), 6, 5
    )


async def test_reminder_cancelled_by_entities(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    tracker_with_six_scheduled_reminders: DialogueStateTracker,
):
    tracker = tracker_with_six_scheduled_reminders
    await default_processor._schedule_reminders(
        tracker.events, tracker, default_channel, default_processor.nlg
    )

    # cancel the fourth reminder
    await cancel_reminder_and_check(
        tracker,
        default_processor,
        ReminderCancelled(entities=[{"entity": "name", "value": "Bruce Wayne"}]),
        6,
        5,
    )


async def test_reminder_cancelled_by_intent(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    tracker_with_six_scheduled_reminders: DialogueStateTracker,
):
    tracker = tracker_with_six_scheduled_reminders
    await default_processor._schedule_reminders(
        tracker.events, tracker, default_channel, default_processor.nlg
    )

    # cancel the third, fifth, and sixth reminder
    await cancel_reminder_and_check(
        tracker, default_processor, ReminderCancelled(intent="default"), 6, 3
    )


async def test_reminder_cancelled_all(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    tracker_with_six_scheduled_reminders: DialogueStateTracker,
):
    tracker = tracker_with_six_scheduled_reminders
    await default_processor._schedule_reminders(
        tracker.events, tracker, default_channel, default_processor.nlg
    )

    # cancel all reminders
    await cancel_reminder_and_check(
        tracker, default_processor, ReminderCancelled(), 6, 0
    )


async def test_reminder_restart(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = uuid.uuid4().hex

    reminder = ReminderScheduled(
        "utter_greet", datetime.datetime.now(), kill_on_user_message=False
    )
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    tracker.update(reminder)
    tracker.update(Restarted())  # cancels the reminder
    tracker.update(UserUttered("test"))

    default_processor.tracker_store.save(tracker)
    await default_processor.handle_reminder(
        reminder, sender_id, default_channel, default_processor.nlg
    )

    # retrieve the updated tracker
    t = default_processor.tracker_store.retrieve(sender_id)
    assert len(t.events) == 4  # nothing should have been executed


@pytest.mark.parametrize(
    "event_to_apply,session_expiration_time_in_minutes,has_expired",
    [
        # last user event is way in the past
        (UserUttered(timestamp=1), 60, True),
        # user event are very recent
        (UserUttered("hello", timestamp=time.time()), 120, False),
        # there is user event
        (ActionExecuted(ACTION_LISTEN_NAME, timestamp=time.time()), 60, False),
        # Old event, but sessions are disabled
        (UserUttered("hello", timestamp=1), 0, False),
        # there is no event
        (None, 1, False),
    ],
)
async def test_has_session_expired(
    event_to_apply: Optional[Event],
    session_expiration_time_in_minutes: float,
    has_expired: bool,
    default_processor: MessageProcessor,
):
    sender_id = uuid.uuid4().hex

    default_processor.domain.session_config = SessionConfig(
        session_expiration_time_in_minutes, True
    )
    # create new tracker without events
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)
    tracker.events.clear()

    # apply desired event
    if event_to_apply:
        tracker.update(event_to_apply)

    # noinspection PyProtectedMember
    assert default_processor._has_session_expired(tracker) == has_expired


# noinspection PyProtectedMember
async def test_update_tracker_session(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    monkeypatch: MonkeyPatch,
):
    sender_id = uuid.uuid4().hex
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    # patch `_has_session_expired()` so the `_update_tracker_session()` call actually
    # does something
    monkeypatch.setattr(default_processor, "_has_session_expired", lambda _: True)

    await default_processor._update_tracker_session(tracker, default_channel)

    # the save is not called in _update_tracker_session()
    default_processor._save_tracker(tracker)

    # inspect tracker and make sure all events are present
    tracker = default_processor.tracker_store.retrieve(sender_id)

    assert list(tracker.events) == [
        ActionExecuted(ACTION_LISTEN_NAME),
        ActionExecuted(ACTION_SESSION_START_NAME),
        SessionStarted(),
        ActionExecuted(ACTION_LISTEN_NAME),
    ]


# noinspection PyProtectedMember
async def test_update_tracker_session_with_slots(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    monkeypatch: MonkeyPatch,
):
    sender_id = uuid.uuid4().hex
    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    # apply a user uttered and five slots
    user_event = UserUttered("some utterance")
    tracker.update(user_event)

    slot_set_events = [SlotSet(f"slot key {i}", f"test value {i}") for i in range(5)]

    for event in slot_set_events:
        tracker.update(event)

    # patch `_has_session_expired()` so the `_update_tracker_session()` call actually
    # does something
    monkeypatch.setattr(default_processor, "_has_session_expired", lambda _: True)

    await default_processor._update_tracker_session(tracker, default_channel)

    # the save is not called in _update_tracker_session()
    default_processor._save_tracker(tracker)

    # inspect tracker and make sure all events are present
    tracker = default_processor.tracker_store.retrieve(sender_id)
    events = list(tracker.events)

    # the first three events should be up to the user utterance
    assert events[:2] == [ActionExecuted(ACTION_LISTEN_NAME), user_event]

    # next come the five slots
    assert events[2:7] == slot_set_events

    # the next two events are the session start sequence
    assert events[7:9] == [ActionExecuted(ACTION_SESSION_START_NAME), SessionStarted()]

    # the five slots should be reapplied
    assert events[9:14] == slot_set_events

    # finally an action listen, this should also be the last event
    assert events[14] == events[-1] == ActionExecuted(ACTION_LISTEN_NAME)


# noinspection PyProtectedMember
async def test_get_tracker_with_session_start(
    default_channel: CollectingOutputChannel, default_processor: MessageProcessor
):
    sender_id = uuid.uuid4().hex
    tracker = await default_processor.get_tracker_with_session_start(
        sender_id, default_channel
    )

    # ensure session start sequence is present
    assert list(tracker.events) == [
        ActionExecuted(ACTION_SESSION_START_NAME),
        SessionStarted(),
        ActionExecuted(ACTION_LISTEN_NAME),
    ]


async def test_handle_message_with_session_start(
    default_channel: CollectingOutputChannel,
    default_processor: MessageProcessor,
    monkeypatch: MonkeyPatch,
):
    sender_id = uuid.uuid4().hex

    entity = "name"
    slot_1 = {entity: "Core"}
    await default_processor.handle_message(
        UserMessage(f"/greet{json.dumps(slot_1)}", default_channel, sender_id)
    )

    assert default_channel.latest_output() == {
        "recipient_id": sender_id,
        "text": "hey there Core!",
    }

    # patch processor so a session start is triggered
    monkeypatch.setattr(default_processor, "_has_session_expired", lambda _: True)

    slot_2 = {entity: "post-session start hello"}
    # handle a new message
    await default_processor.handle_message(
        UserMessage(f"/greet{json.dumps(slot_2)}", default_channel, sender_id)
    )

    tracker = default_processor.tracker_store.get_or_create_tracker(sender_id)

    # make sure the sequence of events is as expected
    assert list(tracker.events) == [
        ActionExecuted(ACTION_SESSION_START_NAME),
        SessionStarted(),
        ActionExecuted(ACTION_LISTEN_NAME),
        UserUttered(
            f"/greet{json.dumps(slot_1)}",
            {"name": "greet", "confidence": 1.0},
            [{"entity": entity, "start": 6, "end": 22, "value": "Core"}],
        ),
        SlotSet(entity, slot_1[entity]),
        ActionExecuted("utter_greet"),
        BotUttered("hey there Core!"),
        ActionExecuted(ACTION_LISTEN_NAME),
        ActionExecuted(ACTION_SESSION_START_NAME),
        SessionStarted(),
        # the initial SlotSet is reapplied after the SessionStarted sequence
        SlotSet(entity, slot_1[entity]),
        ActionExecuted(ACTION_LISTEN_NAME),
        UserUttered(
            f"/greet{json.dumps(slot_2)}",
            {"name": "greet", "confidence": 1.0},
            [
                {
                    "entity": entity,
                    "start": 6,
                    "end": 42,
                    "value": "post-session start hello",
                }
            ],
        ),
        SlotSet(entity, slot_2[entity]),
        ActionExecuted(ACTION_LISTEN_NAME),
    ]


# noinspection PyProtectedMember
@pytest.mark.parametrize(
    "action_name, should_predict_another_action",
    [
        (ACTION_LISTEN_NAME, False),
        (ACTION_SESSION_START_NAME, False),
        ("utter_greet", True),
    ],
)
async def test_should_predict_another_action(
    default_processor: MessageProcessor,
    action_name: Text,
    should_predict_another_action: bool,
):
    assert (
        default_processor.should_predict_another_action(action_name)
        == should_predict_another_action
    )



import numpy as np
import typing
from typing import Any, List, Text, Optional, Dict, Type

from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.components import Component
from rasa.nlu.featurizers.featurizer import DenseFeaturizer
from rasa.nlu.tokenizers.tokenizer import Token, Tokenizer
from rasa.nlu.utils.mitie_utils import MitieNLP
from rasa.nlu.training_data import Message, TrainingData
from rasa.nlu.constants import (
    TEXT,
    TOKENS_NAMES,
    DENSE_FEATURE_NAMES,
    DENSE_FEATURIZABLE_ATTRIBUTES,
)
from rasa.utils.tensorflow.constants import MEAN_POOLING, POOLING

if typing.TYPE_CHECKING:
    import mitie


class MitieFeaturizer(DenseFeaturizer):
    @classmethod
    def required_components(cls) -> List[Type[Component]]:
        return [MitieNLP, Tokenizer]

    defaults = {
        # Specify what pooling operation should be used to calculate the vector of
        # the CLS token. Available options: 'mean' and 'max'
        POOLING: MEAN_POOLING
    }

    def __init__(self, component_config: Optional[Dict[Text, Any]] = None) -> None:
        super().__init__(component_config)

        self.pooling_operation = self.component_config["pooling"]

    @classmethod
    def required_packages(cls) -> List[Text]:
        return ["mitie", "numpy"]

    def ndim(self, feature_extractor: "mitie.total_word_feature_extractor") -> int:

        return feature_extractor.num_dimensions

    def get_tokens_by_attribute(self, example: Message, attribute: Text) -> Any:
        return example.get(TOKENS_NAMES[attribute])

    def train(
        self,
        training_data: TrainingData,
        config: Optional[RasaNLUModelConfig] = None,
        **kwargs: Any,
    ) -> None:

        mitie_feature_extractor = self._mitie_feature_extractor(**kwargs)
        for example in training_data.intent_examples:
            for attribute in DENSE_FEATURIZABLE_ATTRIBUTES:
                self.process_training_example(
                    example, attribute, mitie_feature_extractor
                )

    def process_training_example(
        self, example: Message, attribute: Text, mitie_feature_extractor: Any
    ):
        attribute_tokens = self.get_tokens_by_attribute(example, attribute)
        if attribute_tokens is not None:
            features = self.features_for_tokens(
                attribute_tokens, mitie_feature_extractor
            )
            example.set(
                DENSE_FEATURE_NAMES[attribute],
                self._combine_with_existing_dense_features(
                    example, features, DENSE_FEATURE_NAMES[attribute]
                ),
            )

    def process(self, message: Message, **kwargs: Any) -> None:

        mitie_feature_extractor = self._mitie_feature_extractor(**kwargs)
        features = self.features_for_tokens(
            message.get(TOKENS_NAMES[TEXT]), mitie_feature_extractor
        )
        message.set(
            DENSE_FEATURE_NAMES[TEXT],
            self._combine_with_existing_dense_features(
                message, features, DENSE_FEATURE_NAMES[TEXT]
            ),
        )

    def _mitie_feature_extractor(self, **kwargs) -> Any:
        mitie_feature_extractor = kwargs.get("mitie_feature_extractor")
        if not mitie_feature_extractor:
            raise Exception(
                "Failed to train 'MitieFeaturizer'. "
                "Missing a proper MITIE feature extractor. "
                "Make sure this component is preceded by "
                "the 'MitieNLP' component in the pipeline "
                "configuration."
            )
        return mitie_feature_extractor

    def features_for_tokens(
        self,
        tokens: List[Token],
        feature_extractor: "mitie.total_word_feature_extractor",
    ) -> np.ndarray:

        # remove CLS token from tokens
        tokens_without_cls = tokens[:-1]

        # calculate features
        features = []
        for token in tokens_without_cls:
            features.append(feature_extractor.get_feature_vector(token.text))
        features = np.array(features)

        cls_token_vec = self._calculate_cls_vector(features, self.pooling_operation)
        features = np.concatenate([features, cls_token_vec])

        return features



import json

import logging
import requests
from requests import Response
from sanic import Blueprint, response
from sanic.request import Request
from typing import Text, Dict, Any, List, Callable, Awaitable, Optional

from rasa.constants import DOCS_URL_CONNECTORS
from rasa.core.channels.channel import UserMessage, OutputChannel, InputChannel
from sanic.response import HTTPResponse

from rasa.utils.common import raise_warning

logger = logging.getLogger(__name__)


class MattermostBot(OutputChannel):
    """A Mattermost communication channel"""

    @classmethod
    def name(cls) -> Text:
        return "mattermost"

    @classmethod
    def token_from_login(cls, url: Text, user: Text, password: Text) -> Optional[Text]:
        """Retrieve access token for mattermost user."""

        data = {"login_id": user, "password": password}
        r = requests.post(url + "/users/login", data=json.dumps(data))
        if r.status_code == 200:
            return r.headers["Token"]
        else:
            logger.error(f"Failed to login mattermost user {user}. Response: {r}")
            return None

    def __init__(
        self, url: Text, token: Text, bot_channel: Text, webhook_url: Optional[Text],
    ) -> None:
        self.url = url
        self.token = token
        self.bot_channel = bot_channel
        self.webhook_url = webhook_url

        super(MattermostBot, self).__init__()

    def _post_message_to_channel(self, channel_id: Text, message: Text):
        return self._post_data_to_channel(
            {"channel_id": channel_id, "message": message}
        )

    def _post_data_to_channel(self, data) -> Response:
        """Send a message to a mattermost channel."""

        headers = {"Authorization": "Bearer " + self.token}
        r = requests.post(self.url + "/posts", headers=headers, data=json.dumps(data))
        if not r.status_code == 200:
            logger.error(
                f"Failed to send message to mattermost channel "
                f"{data.get('channel_id')}. Response: {r}"
            )
        return r

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        for message_part in text.strip().split("\n\n"):
            self._post_message_to_channel(self.bot_channel, message_part)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        json_message.setdefault("channel_id", self.bot_channel)
        json_message.setdefault("message", "")

        self._post_data_to_channel(json_message)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image."""

        self._post_data_to_channel(
            {
                "channel_id": self.bot_channel,
                "props": {"attachments": [{"image_url": image}]},
            }
        )

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons to the output."""

        # buttons are a list of objects: [(option_name, payload)]
        # See https://docs.mattermost.com/developer/interactive-messages.html#message-buttons

        actions = [
            {
                "name": button["title"],
                "integration": {
                    "url": self.webhook_url,
                    "context": {"action": button["payload"]},
                },
            }
            for button in buttons
        ]

        props = {"attachments": [{"actions": actions}]}

        self._post_data_to_channel(
            {"channel_id": self.bot_channel, "message": text, "props": props}
        )


class MattermostInput(InputChannel):
    """Mattermost input channel implemenation."""

    @classmethod
    def name(cls) -> Text:
        return "mattermost"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if credentials is None:
            cls.raise_missing_credentials_exception()

        # pytype: disable=attribute-error
        if credentials.get("pw") is not None or credentials.get("user") is not None:
            raise_warning(
                "Mattermost recently switched to bot accounts. 'user' and 'pw' "
                "should not be used anymore, you should rather convert your "
                "account to a bot account and use a token. Password based "
                "authentication will be removed in Rasa Open Source 2.0.",
                FutureWarning,
                docs=DOCS_URL_CONNECTORS + "mattermost/",
            )
            token = MattermostBot.token_from_login(
                credentials.get("url"), credentials.get("user"), credentials.get("pw")
            )
        else:
            token = credentials.get("token")

        return cls(credentials.get("url"), token, credentials.get("webhook_url"),)
        # pytype: enable=attribute-error

    def __init__(self, url: Text, token: Text, webhook_url: Text) -> None:
        """Create a Mattermost input channel.
        Needs a couple of settings to properly authenticate and validate
        messages.

        Args:
            url: Your Mattermost team url including /v4 example
                https://mysite.example.com/api/v4
            token: Your mattermost bot token
            webhook_url: The mattermost callback url as specified
                in the outgoing webhooks in mattermost example
                https://mysite.example.com/webhooks/mattermost/webhook
        """
        self.url = url
        self.token = token
        self.webhook_url = webhook_url

    async def message_with_trigger_word(
        self,
        on_new_message: Callable[[UserMessage], Awaitable[None]],
        output: Dict[Text, Any],
        metadata: Optional[Dict],
    ) -> None:
        # splitting to get rid of the @botmention
        # trigger we are using for this
        split_message = output["text"].split(" ", 1)
        if len(split_message) >= 2:
            message = split_message[1]
        else:
            message = output["text"]

        await self._handle_message(
            message, output["user_id"], output["channel_id"], metadata, on_new_message
        )

    async def action_from_button(
        self,
        on_new_message: Callable[[UserMessage], Awaitable[None]],
        output: Dict[Text, Any],
        metadata: Optional[Dict],
    ) -> None:
        # get the action, the buttons triggers
        action = output["context"]["action"]

        await self._handle_message(
            action, output["user_id"], output["channel_id"], metadata, on_new_message
        )

    async def _handle_message(
        self,
        message: Text,
        sender_id: Text,
        bot_channel: Text,
        metadata: Optional[Dict],
        on_new_message: Callable[[UserMessage], Awaitable[None]],
    ):
        try:
            out_channel = MattermostBot(
                self.url, self.token, bot_channel, self.webhook_url,
            )
            user_msg = UserMessage(
                message,
                out_channel,
                sender_id,
                input_channel=self.name(),
                metadata=metadata,
            )
            await on_new_message(user_msg)
        except Exception as e:
            logger.error(f"Exception when trying to handle message.{e}")
            logger.debug(e, exc_info=True)

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[None]]
    ) -> Blueprint:
        mattermost_webhook = Blueprint("mattermost_webhook", __name__)

        @mattermost_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @mattermost_webhook.route("/webhook", methods=["POST"])
        async def webhook(request: Request) -> HTTPResponse:
            output = request.json

            if not output:
                return response.text("")

            metadata = self.get_metadata(request)
            # handle normal message with trigger_word
            if "trigger_word" in output:
                await self.message_with_trigger_word(on_new_message, output, metadata)

            # handle context actions from buttons
            elif "context" in output:
                await self.action_from_button(on_new_message, output, metadata)

            return response.text("success")

        return mattermost_webhook



import logging
from typing import Any, Dict, Optional, Text

from rasa.constants import DOCS_URL_MIGRATION_GUIDE
from rasa.core.constants import DEFAULT_POLICY_PRIORITY, DIALOGUE
from rasa.core.featurizers import TrackerFeaturizer
from rasa.core.policies.ted_policy import TEDPolicy
from rasa.utils.tensorflow.constants import (
    LABEL,
    HIDDEN_LAYERS_SIZES,
    TRANSFORMER_SIZE,
    NUM_TRANSFORMER_LAYERS,
    NUM_HEADS,
    BATCH_SIZES,
    BATCH_STRATEGY,
    EPOCHS,
    RANDOM_SEED,
    RANKING_LENGTH,
    LOSS_TYPE,
    SIMILARITY_TYPE,
    NUM_NEG,
    EVAL_NUM_EXAMPLES,
    EVAL_NUM_EPOCHS,
    NEGATIVE_MARGIN_SCALE,
    REGULARIZATION_CONSTANT,
    SCALE_LOSS,
    USE_MAX_NEG_SIM,
    MAX_NEG_SIM,
    MAX_POS_SIM,
    EMBEDDING_DIMENSION,
    DROP_RATE_DIALOGUE,
    DROP_RATE_LABEL,
    DROP_RATE_ATTENTION,
    WEIGHT_SPARSITY,
    KEY_RELATIVE_ATTENTION,
    VALUE_RELATIVE_ATTENTION,
    MAX_RELATIVE_POSITION,
    SOFTMAX,
    AUTO,
    BALANCED,
)
from rasa.utils.tensorflow.models import RasaModel
import rasa.utils.common as common_utils

logger = logging.getLogger(__name__)


class EmbeddingPolicy(TEDPolicy):
    """Transformer Embedding Dialogue (TED) Policy is described in
    https://arxiv.org/abs/1910.00486.

    This policy has a pre-defined architecture, which comprises the
    following steps:
        - concatenate user input (user intent and entities), previous system actions,
          slots and active forms for each time step into an input vector to
          pre-transformer embedding layer;
        - feed it to transformer;
        - apply a dense layer to the output of the transformer to get embeddings of a
          dialogue for each time step;
        - apply a dense layer to create embeddings for system actions for each time
          step;
        - calculate the similarity between the dialogue embedding and embedded system
          actions. This step is based on the StarSpace
          (https://arxiv.org/abs/1709.03856) idea.
    """

    # please make sure to update the docs when changing a default parameter
    defaults = {
        # ## Architecture of the used neural network
        # Hidden layer sizes for layers before the dialogue and label embedding layers.
        # The number of hidden layers is equal to the length of the corresponding
        # list.
        HIDDEN_LAYERS_SIZES: {DIALOGUE: [], LABEL: []},
        # Number of units in transformer
        TRANSFORMER_SIZE: 128,
        # Number of transformer layers
        NUM_TRANSFORMER_LAYERS: 1,
        # If 'True' use key relative embeddings in attention
        KEY_RELATIVE_ATTENTION: False,
        # If 'True' use key relative embeddings in attention
        VALUE_RELATIVE_ATTENTION: False,
        # Max position for relative embeddings
        MAX_RELATIVE_POSITION: None,
        # Number of attention heads in transformer
        NUM_HEADS: 4,
        # ## Training parameters
        # Initial and final batch sizes:
        # Batch size will be linearly increased for each epoch.
        BATCH_SIZES: [8, 32],
        # Strategy used when creating batches.
        # Can be either 'sequence' or 'balanced'.
        BATCH_STRATEGY: BALANCED,
        # Number of epochs to train
        EPOCHS: 1,
        # Set random seed to any 'int' to get reproducible results
        RANDOM_SEED: None,
        # ## Parameters for embeddings
        # Dimension size of embedding vectors
        EMBEDDING_DIMENSION: 20,
        # The number of incorrect labels. The algorithm will minimize
        # their similarity to the user input during training.
        NUM_NEG: 20,
        # Type of similarity measure to use, either 'auto' or 'cosine' or 'inner'.
        SIMILARITY_TYPE: AUTO,
        # The type of the loss function, either 'softmax' or 'margin'.
        LOSS_TYPE: SOFTMAX,
        # Number of top actions to normalize scores for loss type 'softmax'.
        # Set to 0 to turn off normalization.
        RANKING_LENGTH: 10,
        # Indicates how similar the algorithm should try to make embedding vectors
        # for correct labels.
        # Should be 0.0 < ... < 1.0 for 'cosine' similarity type.
        MAX_POS_SIM: 0.8,
        # Maximum negative similarity for incorrect labels.
        # Should be -1.0 < ... < 1.0 for 'cosine' similarity type.
        MAX_NEG_SIM: -0.2,
        # If 'True' the algorithm only minimizes maximum similarity over
        # incorrect intent labels, used only if 'loss_type' is set to 'margin'.
        USE_MAX_NEG_SIM: True,
        # Scale loss inverse proportionally to confidence of correct prediction
        SCALE_LOSS: True,
        # ## Regularization parameters
        # The scale of regularization
        REGULARIZATION_CONSTANT: 0.001,
        # The scale of how important is to minimize the maximum similarity
        # between embeddings of different labels.
        NEGATIVE_MARGIN_SCALE: 0.8,
        # Dropout rate for embedding layers of dialogue features.
        DROP_RATE_DIALOGUE: 0.1,
        # Dropout rate for embedding layers of label, e.g. action, features.
        DROP_RATE_LABEL: 0.0,
        # Dropout rate for attention.
        DROP_RATE_ATTENTION: 0,
        # Sparsity of the weights in dense layers
        WEIGHT_SPARSITY: 0.8,
        # ## Evaluation parameters
        # How often calculate validation accuracy.
        # Small values may hurt performance, e.g. model accuracy.
        EVAL_NUM_EPOCHS: 20,
        # How many examples to use for hold out validation set
        # Large values may hurt performance, e.g. model accuracy.
        EVAL_NUM_EXAMPLES: 0,
    }

    def __init__(
        self,
        featurizer: Optional[TrackerFeaturizer] = None,
        priority: int = DEFAULT_POLICY_PRIORITY,
        max_history: Optional[int] = None,
        model: Optional[RasaModel] = None,
        **kwargs: Dict[Text, Any],
    ) -> None:
        super().__init__(featurizer, priority, max_history, model, **kwargs)

        common_utils.raise_warning(
            f"'EmbeddingPolicy' is deprecated and will be removed in version 2.0. "
            f"Use 'TEDPolicy' instead.",
            category=FutureWarning,
            docs=DOCS_URL_MIGRATION_GUIDE,
        )



import shutil

from subprocess import check_call

from _pytest.tmpdir import TempdirFactory
from typing import Callable
import pytest
import shutil
import os
from _pytest.pytester import Testdir, RunResult

from rasa.utils.io import write_yaml_file


@pytest.fixture
def run(testdir: Testdir) -> Callable[..., RunResult]:
    def do_run(*args):
        args = ["rasa"] + list(args)
        return testdir.run(*args)

    return do_run


@pytest.fixture
def run_with_stdin(testdir: Testdir) -> Callable[..., RunResult]:
    def do_run(*args, stdin):
        args = ["rasa"] + list(args)
        return testdir.run(*args, stdin=stdin)

    return do_run


@pytest.fixture(scope="session")
def init_default_project(tmpdir_factory: TempdirFactory) -> str:
    path = tmpdir_factory.mktemp("agent").strpath
    os.environ["LOG_LEVEL"] = "ERROR"

    check_call(["rasa", "init", "--no-prompt"], cwd=path)

    return path


@pytest.fixture
def run_in_default_project_without_models(testdir: Testdir) -> Callable[..., RunResult]:
    os.environ["LOG_LEVEL"] = "ERROR"

    _set_up_initial_project(testdir)

    def do_run(*args):
        args = ["rasa"] + list(args)
        return testdir.run(*args)

    return do_run


@pytest.fixture
def run_in_default_project(
    testdir: Testdir, init_default_project: str
) -> Callable[..., RunResult]:
    os.environ["LOG_LEVEL"] = "ERROR"

    # makes sure we do not always retrain an initial model for every "new" project
    for file_name in os.listdir(init_default_project):
        full_file_name = os.path.join(init_default_project, file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, str(testdir.tmpdir))
        else:
            shutil.copytree(full_file_name, str(testdir.tmpdir / file_name))

    def do_run(*args):
        args = ["rasa"] + list(args)
        result = testdir.run(*args)
        os.environ["LOG_LEVEL"] = "INFO"
        return result

    return do_run


def _set_up_initial_project(testdir: Testdir):
    # copy initial project files
    testdir.copy_example("rasa/cli/initial_project/actions.py")
    testdir.copy_example("rasa/cli/initial_project/credentials.yml")
    testdir.copy_example("rasa/cli/initial_project/domain.yml")
    testdir.copy_example("rasa/cli/initial_project/endpoints.yml")
    testdir.mkdir("data")
    testdir.copy_example("rasa/cli/initial_project/data")
    testdir.run("mv", "nlu.md", "data/nlu.md")
    testdir.run("mv", "stories.md", "data/stories.md")

    # create a config file
    # for the cli test the resulting model is not important, use components that are
    # fast to train
    write_yaml_file(
        {
            "language": "en",
            "pipeline": [{"name": "KeywordIntentClassifier"}],
            "policies": [
                {"name": "MappingPolicy"},
                {"name": "MemoizationPolicy", "max_history": 5},
            ],
        },
        "config.yml",
    )
